function GCInitFunc(aEvent)
{
	//read directories. w.a for showing duplicated cards in a list when reloading in a messenger
	var abManager = Components.classes["@mozilla.org/abmanager;1"].
								getService(Components.interfaces.nsIAbManager);
	var abs = gGCUtils.enumToArray(abManager.directories, Components.interfaces.nsIAbDirectory);

	var func = function() {
		gGCContactManager.init(aEvent);
		var observerService = Components.classes["@mozilla.org/observer-service;1"].
	  												getService(Components.interfaces.nsIObserverService);
		if (!gGCContactManager.uninstallObserver) {
			gGCContactManager.uninstallObserver = new GCUninstallObserver("{BDD92442-0534-4D6F-A966-BAB7D561D781}", "gmcont\.");
			gGCContactManager.uninstallObserver.observe(null, "app-startup", null);
		}
	}
	setTimeout(func,1000);
}

function GCFinFunc(aEvent)
{
	if (gGCContactManager.uninstallObserver) gGCContactManager.uninstallObserver.observe(null, "quit-application", null);
	gGCContactManager.removeAllCache();
	gGCLogger.close();
}

window.addEventListener("load", GCInitFunc, false);
window.addEventListener("unload", GCFinFunc, false);



function GCContacts(account, passwd, manager, syncNow, serialNo)
{
	//Namespace
	this.atom = "http://www.w3.org/2005/Atom";
	this.openSearch = "http://a9.com/-/spec/opensearch/1.1/";
	this.gContact = "http://schemas.google.com/contact/2008";
	this.gd = "http://schemas.google.com/g/2005";
	
	this.gDataAuthToken = null;
	this.serialNo = serialNo;
	this.abManager = Components.classes["@mozilla.org/abmanager;1"].
								getService(Components.interfaces.nsIAbManager);
	this.pref = Components.classes['@mozilla.org/preferences;1'].
	  						getService(Components.interfaces.nsIPrefBranch);
	this.contacts = new Array();
	this.groups = new Array();
	this.orgContacts = null;
	this.orgGroups = null;
	this.undoQueue = new Array();
	this.account = account;
	this.passwd = passwd;
	this.baseFeedUrl = "https://www.google.com/m8/feeds/contacts/"+encodeURIComponent(account)+"/full";
	this.baseGroupsFeedUrl = "https://www.google.com/m8/feeds/groups/"+encodeURIComponent(account)+"/full";
	this.batchFeedUri = "";
	this.manager = manager;
	this.localeBundle = document.getElementById("GmContBundle");
	this.abURI = this.getAddressBookURI();
	this.dateTolerance = 0;
	this.batchMgr = new GCBatchManager(this);
	//this.queueNum = 0;
	this.backMgr = new GCBackupManager();
	this.syncGroupName = "";
	this.syncGroup = null;
	this.offlineChangedItems = null; //working space for batch sync
	this.busy = false; //now syncing
	this.startIndex = 1;
	this.nextLink = "";
	this.resultsPerPage = nsPreferences.getIntPref("gmcont.results_per_page", 100);
	this.yieldCnt = 0;
	this.yieldWeight = Math.pow(10, nsPreferences.getIntPref("gmcont.yield_weight", 2));
	this.myContactsGroupId = null;
	this.reAuth = false;
	this.requestQueue = [];
	this.recordUndo = true;
	
	gGCLogger.dump("Yield weight: "+this.yieldWeight+"\n");
	if (this.abURI) this.manager.startListening(this.abURI);
	if (syncNow) {
		this.auth();
	}
}

GCContacts.prototype.isGroupMappingEnabled = function()
{
	return (this.syncGroupName != "") && this.syncGroup;
}

GCContacts.prototype.deleteContactEntry = function(contact)
{
	if (contact) {
		if (this.recordUndo) this.undoQueue.unshift(contact);
		delete this.contacts[contact.getId()];
	}
}

GCContacts.prototype.deleteGroupEntry = function(group)
{
	if (group) {
		if (this.recordUndo) this.undoQueue.unshift(group);
		delete this.groups[group.getId()];
	}
}

GCContacts.prototype.getDeletedEntries = function()
{
	var entries = new Array();
	for (var i=0; i<this.undoQueue.length; i++) {
		var entry = this.undoQueue[i];
		entry.loadCache();
		var type = "";
		var label = "";
		if (entry instanceof GCContactEntry) {
			type = "gccontact";
			var addr = entry.getPrimaryAddress();
			var name = entry.getDispName();
			label = name + " <" + addr + ">";
		}	else {
			type = "gcgroup";
			label = entry.getDispName();
		}
				
		var item = {
			type: type,
			label: label,
			index: i
		};
		
		entries.push(item);
	}
	
	return entries;
}

GCContacts.prototype.undoDeletedEntry = function(index)
{
	this.batchMgr.init();
	
	var delTags = ["id", "link", "updated"];
	if (index == -1) {
		var entry;
		while (entry = this.undoQueue.shift()) {
			entry.loadCache();
			for (var i=0; i<delTags.length; i++) {
				var tags = entry.entry.getElementsByTagNameNS(this.atom, delTags[i]);
				for (var j=0; j<tags.length; j++) {
					entry.entry.removeChild(tags[j]);
				}
			}
			entry.entry.removeAttribute("etag");
			this.batchMgr.registerAddEntry(entry);
		}
	} else {
		var entry = this.undoQueue[index];
		if (!entry) return;
		entry.loadCache();
		for (var i=0; i<delTags.length; i++) {
			var tags = entry.entry.getElementsByTagNameNS(this.atom, delTags[i]);
			for (var j=0; j<tags.length; j++) {
				entry.entry.removeChild(tags[j]);
			}
		}
		entry.entry.removeAttribute("etag");
		this.batchMgr.registerAddEntry(entry);
		this.undoQueue.splice(index, 1);
	}
	this.batchMgr.doBatch();
}

GCContacts.prototype.dumpToJSConsole = function(msg)
{
	var consoleService = Components.classes["@mozilla.org/consoleservice;1"].getService(Components.interfaces.nsIConsoleService);
	consoleService.logStringMessage(msg);
}

GCContacts.prototype.unsync = function()
{
	this.pref.setBoolPref("gmcont."+this.account+".sync", false);	
}

GCContacts.prototype.getAddressBook = function()
{
	var abURI = this.getAddressBookURI();
	if (!abURI) return null;
	
	var addrBookDir = this.abManager.getDirectory(abURI);
	return addrBookDir;
}

GCContacts.prototype.getAddressBookURI = function(dontCreateAB)
{
	if (!this.account) return null;
	var abURI = null;
	var abManager = Components.classes["@mozilla.org/abmanager;1"]
                          .getService(Components.interfaces.nsIAbManager);
	try {
		abURI = this.pref.getCharPref("gmcont."+this.account+".aburi");
		if (!abURI) throw "No AddrURI";
		//confirm existance of the address book
		var addrBookDir = abManager.getDirectory(abURI);
		if (!addrBookDir) throw "No AddrURI";
		
		//test the address book whether is valid or not
		addrBookDir.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
		var db = addrBookDir.database;
		if (!db) throw "No AddrBook";
	} catch(e) { // create new address book
		if (dontCreateAB) return null;
		gGCLogger.dump("No address book. Create one.\n");
	 	/*
    	const kLDAPDirectory = 0; // defined in nsDirPrefs.h
			const kPABDirectory  = 2; // defined in nsDirPrefs.h
   	*/
  	abManager.newAddressBook("Google-"+this.account, "", 2);
  	//find address book
  	var iter = gGCUtils.enumToArray(abManager.directories);
  	for (var i=0; i<iter.length; i++) {
      var ab = iter[i];
      if (ab instanceof Components.interfaces.nsIAbDirectory && ab.dirName == "Google-"+this.account) {
      	abURI = ab.URI;
      	this.registerAddressBook(abURI);
      	gGCLogger.dump("AbCreated: "+abURI+"\n");
      	break;
      }
    }
	}
	
	return abURI;
}

GCContacts.prototype.registerAddressBook = function(abURI)
{
	if (abURI) {
		this.abURI = abURI;
		try {
			this.pref.setCharPref("gmcont."+this.account+".aburi", abURI);
		} catch(e) {
			gGCLogger.dump(e);
			return false;
		}
	}
	
	return true;
}

GCContacts.prototype.unregisterAddressBook = function()
{
	//this.pref.clearUserPref("gmcont."+this.account+".aburi");
}

GCContacts.prototype.getOfflineChangedItems = function()
{
	var records = this.manager.getOfflineChanges(this.account, true);
	var abURI = this.getAddressBookURI()
	var addrBookDir = this.getAddressBook();
	if (!addrBookDir) {
		//alert("no addrbook");
		return null;
	}
	
	addrBookDir.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
	var db = addrBookDir.database;
	addrBookDir.QueryInterface(Components.interfaces.nsIAbDirectory);
	var upCards = new Array();
	var delCardsIds = new Array();
	var upLists = new Array();
	var delListsIds = new Array();
	var cardModCheckTable = new Array();
	var cardDelCheckTable = new Array();
	var listModCheckTable = new Array();
	var listDelCheckTable = new Array();
	for (var i=0; i<records.length; i++) {
		var record = records[i];
		if (record.type == "card") { //card
			switch (record.operation) {
				case "ADD":
					var card = db.getCardFromAttribute(addrBookDir, "GmOfflineId", record.id, false);
					if (!card) continue;
					if (!cardModCheckTable[record.id]) {
						upCards.push(card);
						cardModCheckTable[record.id] = true;
					}
					break;
				case "MOD":
					var card = db.getCardFromAttribute(addrBookDir, "GmContactId", record.id, false);
					if (!card) continue;
					if (!cardModCheckTable[record.id]) {
						upCards.push(card);
						cardModCheckTable[record.id] = true;
					}
					break;
				case "DEL":
					if (!cardDelCheckTable[record.id]) {
						delCardsIds.push(record.id);
						cardDelCheckTable[record.id] = true;
					}
					break;
			}
		} else { //mail list
			switch (record.operation) {
				case "ADD": case "MOD":
  				var list = this.abManager.getDirectory(record.id);
					if (!list) continue;
					if (!listModCheckTable[record.id]) {
						upLists.push(list);
						listModCheckTable[record.id] = true;
					}
					break;
				case "DEL":
					if (!listDelCheckTable[record.id]) {
						delListsIds.push(record.id);
						listDelCheckTable[record.id] = true;
					}
					break;
			}
			
		}
	}

	var ret = null;
	if (upCards.length > 0 || delCardsIds.length > 0 ||
			upLists.length > 0 || delListsIds.length >0) {
		ret = {
			upCards: upCards,
			delCardsIds: delCardsIds, 
			upLists: upLists,
			delListsIds: delListsIds
		};
	}
	return ret;
}

GCContacts.prototype.submitOfflineChanges = function()
{
	var abURI = this.getAddressBookURI()
	this.manager.stopListening(abURI, true);
	var addrBookDir = this.getAddressBook();
	if (!addrBookDir) {
		//alert("no addrbook");
		this.manager.startListening(abURI, true);
		return false;
	}
	
	addrBookDir.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
	var db = addrBookDir.database;
	addrBookDir.QueryInterface(Components.interfaces.nsIAbDirectory);

	//find cards
	var items = this.getOfflineChangedItems();
	if (!items) {
		this.manager.startListening(abURI, true);
		return false;
	}
	var cards = items.upCards;
	for (var i=0; i<cards.length; i++) {
		var card = cards[i];
		var contactId = null;
		try {
			contactId = card.getProperty("GmContactId", null);
		} catch(e) {
			contactId = null;
		}
		card.QueryInterface(Components.interfaces.nsIAbCard);
		var contact = null;
		if (contactId) contact = this.contacts[contactId];
		if (contact) {
				contact.setABCard(card);
				contact.updateGContactFromCard(card);
			 	this.batchMgr.registerModEntry(contact);
		}	else {
			contact = new GCContactEntry();
			contact.setABCard(card);
			contact.updateGContactFromCard(card);
			if (this.syncGroup) {
				contact.belongToGroup(this.syncGroup);
			} else if (nsPreferences.getBoolPref("gmcont.add_to_mycontacts", false) && this.myContactsGroupId) {
				contact.belongToGroup(this.myContactsGroupId);
			}
			this.batchMgr.registerAddEntry(contact);
		}
	}

	var lists = items.upLists;
	for (var i=0; i<lists.length; i++) {
		var dir = lists[i];
		var group = this.getGroupByMailList(dir);
		if (group) {
				group.setMailList(dir, this.account);
				group.updateGGroupFromMailList(dir);
				this.batchMgr.registerModEntry(group);
		} else {
			group = new GCGroupEntry();
			group.setMailList(dir, this.account);
			group.updateGGroupFromMailList(dir);
			this.batchMgr.registerAddEntry(group);
		}
	}
	
  this.manager.startListening(abURI, true);
  this.offlineChangedItems = items;
  try {
	return this.batchMgr.doBatch();
  }catch(e){
  	gGCLogger.dump("Error in do batch\n");
  	gGCLogger.dump(e+"\n");
  }
}

GCContacts.prototype.getChildCardsInAB = function(db, addrBookDir, cards, listCards)
{
  var childCards = gGCUtils.enumToArray(addrBookDir.childCards, Components.interfaces.nsIAbCard);
  for (var i=0; i<childCards.length; i++) {
	  var card = childCards[i];
		if (card.isMailList) {
		  if (listCards) listCards.push(card);
		} else {
				if (cards) cards.push(card);
		}
  }
}

GCContacts.prototype.refreshABCards = function(gcGolden)
{
	if (!gcGolden && this.manager.syncMode == this.manager.SYNC_MODE_BATCH && !this.batchMgr.submitted) {
		gGCLogger.dump("Batch\n");
		try {
		if (this.submitOfflineChanges()) return;
		}catch(e){
			gGCLogger.dump("Error in submit offline changes\n");
			gGCLogger.dump(e+"\n");
		}
	} else {
		this.manager.clearBatchLog(this.account); //remove batch log if it exists
		gGCLogger.dump("Submitting offline changes was done.\n");
	}
	
	//yield
	var gen = this.refreshABCardsMain(gcGolden);
	var genFunc = function() {
		try {
			gen.next();
			setTimeout(genFunc, 100);
		} catch (e if e instanceof StopIteration) {
		} catch (e) {
			gGCLogger.dump(e+"\n");
		}
	}
	setTimeout(genFunc, 100);
}

GCContacts.prototype.isDeletedOffline = function(item, isContact)
{	
		if (!this.offlineChangedItems) return false;
		var delItems = isContact 
								 ? this.offlineChangedItems.delCardsIds
								 : this.offlineChangedItems.delListsIds;
		var found = delItems.indexOf(item.getId()) >= 0;
		return found;
}

GCContacts.prototype.needYield = function()
{
	this.yieldCnt += 1;
	var ret = false;
	if (this.yieldCnt % this.yieldWeight == 0) {
		this.yieldCnt = 0;
		ret = true
	}
	
	return ret;
}

GCContacts.prototype.refreshABCardsMain = function(gcGolden)
{	
	gGCLogger.dump("Refresh\n");
	var addrBookDir = this.getAddressBook();
	var abURI = this.getAddressBookURI();
	//addrBookDir.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
	//var db = addrBookDir.database;
	//addrBookDir.QueryInterface(Components.interfaces.nsIAbDirectory);
	var msg = this.localeBundle.getString("gmcont.batch_done");
	this.outputMessageToStatusBar(msg);
	this.manager.progressStart();
	this.manager.stopListening(abURI, true);
	
	//find cards
	var delCards = new Array(); //cards to be deleted from AB
 	var delLists = new Array(); //lists to be deleted from AB

	var uninstallObserver = new GCUninstallObserver();
	
 	this.batchMgr.init();
  //this.getChildCardsInAB(db, addrBookDir, delCards, delLists);

	var newContacts = [];
	var newGroups = [];
	var childCards = gGCUtils.enumToArray(addrBookDir.childCards, Components.interfaces.nsIAbCard);
  var len = childCards.length;
  for (var i=0; i<len; i++) {
	  var card = childCards[i];
		if (card.isMailList) { //For list
			var groupId = nsPreferences.copyUnicharPref("gmcont."+this.account+"."+card.mailListURI+".groupId", null);
			var group = groupId ? this.groups[groupId] : null;
			if (group) {
				var list = this.abManager.getDirectory(card.mailListURI);
				group.setMailList(list);
				group.setListCard(card);
				newGroups[groupId] = group;
				delete this.groups[groupId];
			} else {
				delLists.push(card);
			}
		} else { //For card
			var contactId = card.getProperty("GmContactId", null);
			var contact = contactId ? this.contacts[contactId] : null;
			if (contact) {
				try{
					if (contact.updateCardFromGContact(card)) addrBookDir.modifyCard(card);
				}catch(e){
					gGCLogger.dump("Error in updating a card from a contact\n");
					gGCLogger.dump(e+"\n");
				}
				contact.setABCard(card);
				newContacts[contactId] = contact;
				delete this.contacts[contactId]; //already checked;
			} else {
				delCards.push(card);
			}
		}
		if (this.needYield()) yield;
  }

	//check offline deleted
	var listKey ="";
	for (listKey in this.groups) {
		var group = this.groups[listKey];
		if (!group) continue;
		if (this.isDeletedOffline(group, false)) { //deleted offline
    	gGCLogger.dump("List deleted offline\n");
    	this.batchMgr.registerDelEntry(group);
		} else { //new group
			newGroups[listKey] = group;
		}
	}

	this.groups = newGroups; //update groups
	
	if (this.needYield()) yield;
		
	//generate cards
	var delMapMode = nsPreferences.getIntPref("gmcont.map_del_mode",0);
	if (!this.syncGroup) delMapMode = null; //group mapping is disabled
	var key = "";
	for (key in this.contacts) {
		var contact = this.contacts[key];
		if (!contact) continue;
    if (this.isDeletedOffline(contact, true)) { //deleted offline
    	if (delMapMode == 2) delMapMode = this.askDelOperationForGroupMapping();
    	if (delMapMode == 1) {
    		contact.unbelongToGroup(this.syncGroup);
    		contact.map_del_flag = true;
    		this.batchMgr.registerModEntry(contact);
    	} else {
    		this.batchMgr.registerDelEntry(contact);
    	}
    } else { //new contact
    	var card = contact.createABCard(addrBookDir);
			try{
				contact.updateCardFromGContact(card);
			}catch(e){
				gGCLogger.dump("Error in updating a card for a added contact\n");
				gGCLogger.dump(e+"\n");
			}
	  	card.setProperty("GmBookUri", abURI);
	  	card.setProperty("GmContactId", key);
			addrBookDir.modifyCard(card);
    	contact.setABCard(card);
			newContacts[key] = contact;
    }

    if (this.needYield()) yield;
	}
	
	this.contacts = newContacts;
	
	var addItems = null;
 	if (!gcGolden
 			&& this.manager.syncMode == this.manager.SYNC_MODE_BATCH 
 			&& nsPreferences.getBoolPref("gmcont.confirm_batchdel", true)) {
  	this.batchMgr.confirmDeletion();
 		addItems = this.confirmDeletionFromTB(delCards, delLists);
 	}

  if (this.needYield()) yield;
  
 	if (addItems) {
		//add contacts/groups to Google (offline sync)
 		len = addItems.cards.length;
		for (var i=0; i<len; i++) {
			var card = addItems.cards[i];
			var contact = new GCContactEntry();
			contact.setABCard(card);
			contact.updateGContactFromCard(card);
			this.batchMgr.registerAddEntry(contact);
			if (this.needYield()) yield;
		}
		len = addItems.lists.length;
		for (var i=0; i<len; i++) {
			var group = new GCGroupEntry();
			var dir = this.abManager.getDirectory(addItems.lists[i].mailListURI);
			group.setMailList(dir, this.account);
			group.updateGGroupFromMailList(dir);
			this.batchMgr.registerAddEntry(group);
			if (this.needYield()) yield;
		}
	}
	
	if (this.needYield()) yield;
	
		//submit offline deletions and user specified addtions
	this.offlineChangedItems = null;
	if (this.batchMgr.doBatch()) {
		this.manager.startListening(abURI, true);
		this.manager.progressEnd();
		return; //need to do batch. refresh again.
	}

	//delete all list
	var delCardsMutableArray = Components.classes["@mozilla.org/array;1"].createInstance(Components.interfaces.nsIMutableArray);
	//delete unnecessary mail lists
	len=delLists.length;
	for (var i=0; i<len; i++) {
		var delList = delLists[i];
		if (delList) {
			gGCLogger.dump("delete: "+delList.displayName+"\n");
			delCardsMutableArray.appendElement(delList, false);
		}
		if (this.needYield()) yield;
	}
	if (delCardsMutableArray.length > 0) addrBookDir.deleteCards(delCardsMutableArray);

	if (this.needYield()) yield;
	
	//clear last modified date for mail lists
	uninstallObserver.clearPrefs('gmcont\..*\.last_modified');

	//clear list all org names
 	uninstallObserver.clearPrefs('gmcont\.'+this.account+'\..*\.org_name');
 	uninstallObserver.clearPrefs('gmcont\.'+this.account+'\..*\.groupId');
 	
	//create list
	gGCLogger.dump("create list\n");
	var groupKey;
	for (groupKey in this.groups) {
		var group = this.groups[groupKey];
		if (!group) continue;
		var name = group.getTitle();
		if (group.isSystemGroup() && this.serialNo) {
			group.systemDispName = name.split(":")[0] + ":" + this.serialNo;
			name = group.getTitle();
		} else if (!group.isSystemGroup() && this.mailListNameExists(name, group.getMailList())) {
			group.setSuffix(":"+this.serialNo);
		}
		if (group.getMailList()) {
			group.clearMailListChildren();
			gGCLogger.dump("Reuse list for "+group.getTitle()+"\n");
			group.setMailList(group.getMailList());
		} else {
			gGCLogger.dump("Create list: "+group.getTitle()+"\n");
			var list = group.createMailList(addrBookDir);
			addrBookDir.addMailList(list);
			group.setMailList(list);
		}
		group.updateMailListFromGGroup(group.getMailList());
	}
	
	yield;
		
	//delete unnecessary cards
	delCardsMutableArray = Components.classes["@mozilla.org/array;1"].createInstance(Components.interfaces.nsIMutableArray);
	len = delCards.length;
	for (var i=0; i<len; i++) {
		var delCard = delCards[i];
		if (!delCard) continue;
		delCard.QueryInterface(Components.interfaces.nsIAbCard);
		gGCLogger.dump("delete: "+delCard.getProperty("PrimaryEmail", "")+"\n");
		delCardsMutableArray.appendElement(delCard, false);
		if (this.needYield()) yield;
	}

	if (delCardsMutableArray.length > 0) addrBookDir.deleteCards(delCardsMutableArray);

	yield;
		
	var key = null;
	for (key in this.contacts) {
		var contact = this.contacts[key];
		if (!contact || !contact.getPrimaryAddress()) {
			gGCLogger.dump("Does not have a primary address\n");
			continue;
		}
		var card = contact.getABCard();
		if (!card) {
			gGCLogger.dump("Card is not set for:"+contact.getPrimaryAddress()+"\n");
			continue; //deleted offline card
		}
		card.QueryInterface(Components.interfaces.nsIAbCard);
		var groups = contact.getBelongingGroupIds();
		len=groups.length;
		for (var i=0; i<len; i++) {
			var groupId = groups[i];
			var group = this.groups[groupId];
			if (!group) continue; //removed but reference is remained in contact
			var list = group.getMailList();
			if (list) {
				list.addressLists.appendElement(card, false);
				//group.setMailList(list, this.account);
			}
			if (this.needYield()) yield;
		}
		if (this.needYield()) yield;
	}

  yield;

 	var childCards = null;
  try {
		var searchQuery = "?(not(IsMailList,=,FALSE))";
		var resultDir = this.abManager.getDirectory(abURI + searchQuery);
		childCards = resultDir ? resultDir.childCards : null;
  } catch(e) {
  	gGCLogger.dump(e+"\n");
  	childCards = null;
  }
  
  var childCardsArray = gGCUtils.enumToArray(childCards, Components.interfaces.nsIAbCard);
  len = childCardsArray.length;
  for (var i=0; i<len; i++) {
	  var list = childCardsArray[i];
		try {
			var dir = this.abManager.getDirectory(list.mailListURI);
			var group = this.getGroupByMailList(dir);
			list.setProperty("DisplayName", dir.dirName);
      list.lastName = dir.dirName;
      list.setProperty("NickName", dir.listNickName);
      list.setProperty("Notes", dir.description); 
      dir.editMailListToDatabase(list);
			dir = this.abManager.getDirectory(list.mailListURI);
			group.setMailList(dir, this.account);
			group.setListCard(list);
		} catch(e) {
			gGCLogger.dump(e+"\n");
		}
		if (this.needYield()) yield;
  }

	//this.updateLastSyncDate();
	this.manager.startListening(abURI, true);
	//this.manager.progressEnd();
	//this.outputMessageToStatusBar("All synchronization processes are done.");
	
	var gen = this.swapOutMemCache();
	var me = this;
	var genFunc = function() {
		try {
			gen.next();
			setTimeout(genFunc, 100);
		} catch (e if e instanceof StopIteration) {
			me.manager.progressEnd();
			var msg = me.localeBundle.getString("gmcont.all_done");
			msg = msg.replace("%S", me.account);
			me.outputMessageToStatusBar(msg);
			me.busy = false;
			if (nsPreferences.getBoolPref("gmcont.always_reload_photos", false)) {
				me.retrievePhotoAll();
			} else {
				me.manager.notifySyncDone();
			}
		} catch (e) {
			gGCLogger.dump(e+"\n");
		}
	}
	setTimeout(genFunc, 100);
}

GCContacts.prototype.mailListNameExists = function(name, list)
{
	var abManager = Components.classes["@mozilla.org/abmanager;1"]
                .getService(Components.interfaces.nsIAbManager);
  if (list && name == list.dirName) return false;
  else return abManager.mailListNameExists(name);
}

GCContacts.prototype.askDelOperationForGroupMapping = function()
{
	var windowMediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].
												getService(Components.interfaces.nsIWindowMediator);
	var addressbook = windowMediator.getMostRecentWindow("mail:addressbook");
	var win = addressbook ? addressbook : window;
	var callback = {
		operation: null
	};
	win.openDialog("chrome://gmcont/content/askDelOperationDialog.xul", "gmcont-alert", "chrome,modal,dialog,centerscreen", callback);
	
	return callback.operation;
}

GCContacts.prototype.swapOutMemCache = function()
{
	var key;
	for (key in this.contacts) {
		var contact = this.contacts[key]
		if (!contact.purgeMemCache()) {
			gGCLogger.dump("Cannot purge since disk cache is disabled. The rests are skipped.\n");
			return;
		}
		if (this.needYield()) yield;
	}
	
	yield;
	
	for (key in this.groups) {
		var group = this.groups[key];
		if (!group.purgeMemCache()) {
			gGCLogger.dump("Cannot purge since disk cache is disabled. The rests are skipped.\n");
			return;
		}
		if (this.needYield()) yield;
	}
	gGCLogger.dump("Swap out MemCache Done.\n");
	//this.manager.progressEnd();
}

GCContacts.prototype.confirmDeletionFromTB = function(cards, lists)
{		
	var items = new Array();
	
	for (var i=0; i<cards.length; i++) {
		var card = cards[i];
		if (!card) continue;
		try {
			card.QueryInterface(Components.interfaces.nsIAbCard);
		} catch(e) {
			continue;
		}
		var contactId = card.getProperty("GmContactId", null);
		if (!contactId) continue; //do not confrim cards which are not bound to Google contacts

		var addr = card.getProperty("PrimaryEmail", "");
		var name = card.getProperty("DisplayName", "");
		var label = name + " <" + addr + ">";
		var item = {
			label: label,
			type: "card",
			index: i,
			checked: true
		};
		items.push(item);
	}
	
	for (var i=0; i<lists.length; i++) {
		var item = {
			label: lists[i].getProperty("DisplayName", ""),
			type: "list",
			index: i,
			checked: true
		};
		items.push(item);
	}

	if (items.length == 0) return null;

	//var msg = "The following cards/lists in the Thunderbird address book %1 will be removed. Check contacts you want to delete.";
	var msg = this.localeBundle.getString("gmcont.conf_batchdel_tb");
	msg = msg.replace("%S", this.account);
	var title = this.localeBundle.getString("gmcont.conf_batchdel_title_tb");
	var callback = {
		title: title,
		items: items,
		desc: msg,
		isOK: false,
		rows: 10
	};

	var windowMediator = Components.classes["@mozilla.org/appshell/window-mediator;1"].
												getService(Components.interfaces.nsIWindowMediator);
	var addressbook = windowMediator.getMostRecentWindow("mail:addressbook");
	var win = addressbook ? addressbook : window;
	win.openDialog("chrome://gmcont/content/confirmContactsForm.xul", "gmcont-alert", "chrome,modal,dialog,centerscreen", callback);

	nsPreferences.setBoolPref("gmcont.confirm_batchdel", !callback.notShowAgain);
	var addCards = new Array();
	var addLists = new Array();
	for (var i=0; i<items.length; i++) {
		var item = items[i];
		if (!callback.isOK || !item.checked) {
			if (item.type == "card") {
				addCards.push(cards[item.index]);
			} else {
				addLists.push(lists[item.index]);
			}
		}
	}

	return {cards: addCards, lists: addLists};
}

GCContacts.prototype.popMailListByName = function(name, lists, noReuse)
{
	var noNameListIndex = -1;
	for (var i=0; i<lists.length; i++) {
		var listCard = lists[i];
		if (!listCard) continue;
		listCard.QueryInterface(Components.interfaces.nsIAbCard);
	  //var dir = this.rdf.GetResource(listCard.mailListURI).
  	//					QueryInterface(Components.interfaces.nsIAbDirectory);
		var dir = this.abManager.getDirectory(listCard.mailListURI);
  	var dirName = dir.dirName ? dir.dirName : listCard.displayName;
		if (!dirName) {
			noNameListIndex = i;
			gGCLogger.dump("No list name\n");
			continue;
		}

		if (dirName == name) {
			lists.splice(i,1);
			return listCard;
		}
	}

	if (!noReuse && noNameListIndex != -1) {
		gGCLogger.dump("reuse no name list\n");
		var listCard = lists[noNameListIndex];
		lists.splice(noNameListIndex,1);
		return listCard;
	}
	return null;
}

GCContacts.prototype.popCardById = function(id, cards)
{
	for (var i=0; i<cards.length; i++) {
		var card = cards[i];
		if (!card) continue;
		
		//card.QueryInterface(Components.interfaces.nsIAbMDBCard);
		card.QueryInterface(Components.interfaces.nsIAbCard);
		//var contactId = card.getStringAttribute("GmContactId");
		var contactId = card.getProperty("GmContactId", null);
		if (contactId == id) {
			//cards[i] = null;
			cards.splice(i,1);
			return card;
		}
	}
	
	return null;
}

GCContacts.prototype.auth = function(account, passwd)
{
	if (this.busy) {
		gGCLogger.dump("Busy now. Abort auth\n");
		return false;
	}
	/*
	if (this.gDataAuthToken) {
		gGCLogger.dump("Reuse Auth Token\n");
		this.reload();
		return;
	}
	*/
	var httpRequest = new XMLHttpRequest();
	//if (this.gDataAuthToken) return true;
	if (account) this.account = account;
	if (passwd) this.passwd = passwd;
	if (!this.passwd) { //auth will fails if no passwd is specified
		this.authFailedCallback();
		return false;
	}
	
	var feedUrl = "https://www.google.com/accounts/ClientLogin";
 	httpRequest.open("POST", feedUrl, true);
 	
 	// build request body
 	var appInfo = Components.classes["@mozilla.org/xre/app-info;1"].
	 	              getService(Components.interfaces.nsIXULAppInfo);
  var source = appInfo.vendor + "-" + appInfo.name + "-" + appInfo.version;
 	var postBody = "Email=" + encodeURIComponent(this.account) +
   	             "&Passwd=" + encodeURIComponent(this.passwd) +
     	           "&accountType=HOSTED_OR_GOOGLE" + //Gmail + Google Apps
       	         "&source=" + encodeURIComponent(source) +
         	       "&service=cp"; //Contacts
  var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
	                  createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
 	converter.charset = "UTF-8";
 	var stream = converter.convertToInputStream(postBody);
 	httpRequest.setRequestHeader("Content-Type", "application/x-www-form-urlencoded");
 	//httpRequest.setRequestHeader("GData-Version", "2");
 	httpRequest.setRequestHeader("GData-Version", "3.0");
 	
 	// post process func
 	var me = this;
 	var recvFunc = function() {
		gGCLogger.dump("[Auth] " + httpRequest.status+": "+httpRequest.statusText+"\n");
		me.outputMessageToStatusBar("[Auth] " + httpRequest.status+": "+httpRequest.statusText);
	 	switch (httpRequest.status) {
		  case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
      	var msg = me.localeBundle.getString("gmcont.auth_succeeded");
      	msg = msg.replace(/\%S/,me.account);
      	me.outputMessageToStatusBar(msg);
     		var data = httpRequest.responseText;
     		var offset = data.indexOf("Auth=");
       	me.gDataAuthToken = data.substring(offset+5, data.length - 1);
       	//me.retrieve();
     		me.syncGroupName = nsPreferences.copyUnicharPref("gmcont."+me.account+".sync_group", "");
       	//me.takeSnapShotForTbAb();
       	//me.retrieveGroups();
     		me.reload();
        break;
      case 401: /* Authorization required. */
      case 403: /* Unsupported standard parameter, or authentication or
                   Authorization failed. */
      case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
				me.gDataAuthToken = null;
				me.authFailedCallback();
				break;
      default:
      	me.gDataAuthToken = null;
				me.authFailedCallback();
        break;
	 	}
 	}
 	this.sendHttpRequest(httpRequest, stream, recvFunc);
 	
 	return true;
}

GCContacts.prototype.authFailedCallback = function()
{
	gGCLogger.dump("auth again\n");
  var watcher = Components.classes["@mozilla.org/embedcomp/window-watcher;1"].
	               getService(Components.interfaces.nsIWindowWatcher);
  var prompter = watcher.getNewPrompter(null);
	var unameObj = {value: this.account};
	var passwordObj = {value: null};
	var savepasswordObj = {value: false};
	var title = this.localeBundle.getString("gmcont.passwd_title");
	var text = this.localeBundle.getString("gmcont.passwd_text");
	text = text.replace(/\%S/, this.account);
	var savePasswordText = this.localeBundle.getString("gmcont.passwd_save");
	var ret = prompter.promptPassword(title,
	                   								text,
	                   								passwordObj,
	                   								savePasswordText,
	                   								savepasswordObj);
	if (!ret) {  //canceled
   	var msg = this.localeBundle.getString("gmcont.auth_failed");
   	msg = msg.replace(/\%S/, this.account);
   	this.outputMessageToStatusBar(msg);
		this.account = null;
		this.passwd = null;
		this.manager.notifyAuthFailed(this);
		return;
	}
	var password = passwordObj.value;
	if (savepasswordObj.value) {
	  try {
		  var passwordmanager = Components.classes["@mozilla.org/login-manager;1"]
	                        .getService(Components.interfaces.nsILoginManager);
	
	    //remove old password.
	    var logins = passwordmanager.getAllLogins({});
	    for (var i = 0; i < logins.length; i++) {
	      if (logins[i].username != this.account) continue;
	      if (logins[i].hostname == 'chrome://gmcont/'+this.account) { 
	         passwordmanager.removeLogin(logins[i]);
		    }
	   	}
	   	
	    var nsLoginInfo = new Components.Constructor("@mozilla.org/login-manager/loginInfo;1",
	                                             Components.interfaces.nsILoginInfo,
	                                             "init");
			var loginInfo = new nsLoginInfo('chrome://gmcont/'+this.account,
	                      'User Registration', null,
	                      this.account, password, "", "");
	
			//passwordmanager.addUser(this.account, this.account, password);
	    passwordmanager.addLogin(loginInfo);
	  } catch(e) {
	  	gGCLogger.dump(e);
	  }
	}
	this.passwd = password;
	this.auth(this.account, this.passwd);
}

GCContacts.prototype.reload = function()
{
	this.takeSnapShotForTbAb();
	this.orgGroups = this.groups;
	this.orgContacts = this.contacts;
	this.groups = [];
	this.contacts = [];
	this.retrieveGroups();
}

// get all contacs
GCContacts.prototype.retrieve = function(id)
{
	if (!this.gDataAuthToken) {
		//alert(this.localeBundle.getString("gmcont.auth_needed"));
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		return;
	}
	//feed Url for retrieving all contacts
	var feedUrl = this.baseFeedUrl;
	if (id) {
		var contact = this.contacts[id];
		if (!contact) {
			gGCLogger.dump("Error: No such contact.\n")
			return;
		}
		//feed Url for the contact
		feedUrl = contact.getSelfLink();
	} else {
		//var maxResults = nsPreferences.getIntPref("gmcont.max-results", 0);
		//if (maxResults > 0) feedUrl = feedUrl + "?max-results=" + maxResults;
		//feedUrl = feedUrl + "?max-results=" + this.resultsPerPage +"&start-index=" + this.startIndex;
		//this.startIndex += this.resultsPerPage;
		feedUrl = feedUrl + "?max-results=" + this.resultsPerPage +"&start-index=1";
		if (this.syncGroupName != "" && this.syncGroup) {
			var groupId = this.syncGroup.getId();
			gGCLogger.dump("Get contacts in: "+groupId+"\n");
			feedUrl = feedUrl + "&group=" + groupId;
		}
		if (this.nextLink) feedUrl = this.nextLink;
		this.startIndex += this.resultsPerPage;
	}
	
  var httpRequest = new XMLHttpRequest();
	httpRequest.open("GET", feedUrl, true);
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
	//httpRequest.setRequestHeader("GData-Version", "2");
	httpRequest.setRequestHeader("GData-Version", "3.0");
	//httpRequest.setRequestHeader("If-Match", "*");
	
	var me = this;
  var recvFunc = function() {
   	gGCLogger.dump("[Retrieve] "+httpRequest.status+": "+httpRequest.statusText+"\n");
		me.outputMessageToStatusBar("[Retrieve] " + httpRequest.status+": "+httpRequest.statusText);
   	switch (httpRequest.status) {
	    case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
      	if (id) {
      		me.contacts[id].init(httpRequest.responseXML.documentElement);
   				me.outputMessageToStatusBar("");
      	} else {
					/*
      		var parseFunc = function() {
 	  				me.parseFeed(httpRequest.responseText);
 	  			}
 	  			setTimeout(parseFunc);
					*/
 	  			var gen = me.parseFeedXML(httpRequest.responseXML);
 	  			var genFunc = function() {
						try {
							gen.next();
							setTimeout(genFunc, 100);
						} catch(e if e instanceof StopIteration) {
						} catch(e) {
 	  					gGCLogger.dump(e+"\n");
						}
 	  			}
 	  			setTimeout(genFunc, 100);
      	}
       	break;
      case 304: /* Not Modified */
      	break;
      case 401: /* Authorization required. */
      case 403: /* Unsupported standard parameter, or authentication or
                   Authorization failed. */
        if (me.reAuth) {
        	me.reAuth = false;
        	gGCLogger.dump("Auth problem. Cannot retrieve contacts.\n");
        } else {
        	me.reAuth = true;
        	gGCLogger.dump("Need to re-auth\n");
      		me.gDataAuthToken = null;
      		me.auth();
        }
        break;
      case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
      	break;
      default:
      	break;
    	}
   	};

	 	this.sendHttpRequest(httpRequest, null, recvFunc);
	 	return;
}

// get all contacs
GCContacts.prototype.retrieveGroups = function(id)
{
	if (!this.gDataAuthToken) {
		//alert(this.localeBundle.getString("gmcont.auth_needed"));
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		return;
	}
	//feed Url for retrieving all contacts
	var feedUrl = this.baseGroupsFeedUrl;
	if (id) {
		var group = this.groups[id];
		if (!contact) {
			gGCLogger.dump("Error: No such group.\n")
			return;
		}
		//feed Url for the contact
		feedUrl = group.getSelfLink();
	} else {
		if (this.startIndex == 1) {
			this.contacts = new Array();
			this.groups = new Array();
		}
		//var maxResults = nsPreferences.getIntPref("gmcont.max-results", 0);
		//if (maxResults > 0) feedUrl = feedUrl + "?max-results=" + maxResults;
		//feedUrl = feedUrl + "?max-results=" + this.resultsPerPage +"&start-index=" + this.startIndex;
		//this.startIndex += this.resultsPerPage;
		feedUrl = feedUrl + "?max-results=" + this.resultsPerPage +"&start-index=1";
		if (this.nextLink) feedUrl = this.nextLink;
		this.startIndex += this.resultsPerPage;
	}
	
  var httpRequest = new XMLHttpRequest();
	httpRequest.open("GET", feedUrl, true);
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
	//httpRequest.setRequestHeader("GData-Version", "2");
	httpRequest.setRequestHeader("GData-Version", "3.0");
	//httpRequest.setRequestHeader("If-Match", "*");
	
	var me = this;
  var recvFunc = function() {
   	gGCLogger.dump("[Retrieve Groups] "+httpRequest.status+": "+httpRequest.statusText+"\n");
		me.outputMessageToStatusBar("[Retrieve Groups] " + httpRequest.status+": "+httpRequest.statusText);
   	switch (httpRequest.status) {
	    case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
      	if (id) {
      		me.groups[id].init(httpRequest.responseXML.documentElement);
      		me.groups[id].initSystemGroups();
      	} else {
 	  			/*
 	  			var parseFunc = function() {
 	  				me.parseFeed(httpRequest.responseText, true);
 	  			}
 	  			setTimeout(parseFunc);
 	  			*/
 	  			//var gen = me.parseFeed(httpRequest.responseText, true);
      		var gen = me.parseFeedXML(httpRequest.responseXML, true);
 	  			var genFunc = function() {
 	  				try {	
							gen.next();
							setTimeout(genFunc, 100);
						} catch(e if e instanceof StopIteration) {
 	  					//var msg = me.localeBundle.getString("gmcont.reload_done");
	  					//msg = msg.replace(/\%S/, me.account);
  						//me.outputMessageToStatusBar(msg);
  						//me.retrieve();
						} catch(e) {
							gGCLogger.dump(e+"\n");
						}
 	  			}
 	  			setTimeout(genFunc, 100);
				}
       	break;
      case 304: /* Not Modified */
      	break;
      case 401: /* Authorization required. */
      case 403: /* Unsupported standard parameter, or authentication or
                   Authorization failed. */
        if (me.reAuth) {
        	me.reAuth = false;
        	gGCLogger.dump("Auth problem. Cannot retrieve groups.\n");
        } else {
        	me.reAuth = true;
        	gGCLogger.dump("Need to re-auth\n");
      		me.gDataAuthToken = null;
      		me.auth();
        }
      	break;
      case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
      	break;
      default:
      	break;
    	}
   	};

	 	this.sendHttpRequest(httpRequest, null, recvFunc);
	 	return;
}

GCContacts.prototype.retrievePhotoAll = function()
{
	var photoQueue = new Array();
	var cKey;
	for (cKey in this.contacts) {
		photoQueue.push(this.contacts[cKey]);
	}

	this.retrievePhoto(photoQueue);
}

GCContacts.prototype.retrievePhoto = function(item)
{
	if (!this.gDataAuthToken) {
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		this.manager.notifySyncDone();
		return;
	}
	
	var addrBookURI = this.getAddressBookURI();
	if (this.manager.listening(addrBookURI)) this.manager.stopListening(addrBookURI);
	//this.manager.stopListening(addrBookURI);
	var contact = null;
	if (item instanceof Array) {
		contact = item.pop();
	} else if (item instanceof GCContactEntry) {
		contact = item;
	}

	if (!contact) {
		this.manager.startListening(addrBookURI);
		this.outputMessageToStatusBar("");
		this.manager.notifySyncDone();
		return;
	}

	contact.setPhotoManager(this.manager.photoManager);
	var card = contact.getABCard();
	var addrBookDir = this.getAddressBook();
	var oldEtag = card.getProperty("GmPhotoEtag", null);
	var feedUrl = contact.getPhotoLink();
	if (!feedUrl) { //no photo
		if (oldEtag) {
			card = contact.resetPhoto();
			card.deleteProperty("GmPhotoEtag");
			addrBookDir.modifyCard(card);
		}
		this.outputMessageToStatusBar("");
		if (item instanceof Array) this.retrievePhoto(item);
		contact.clearPhotoManager();
		return;
	}

	var newEtag = contact.getPhotoEtag();

  var httpRequest = new XMLHttpRequest();
	httpRequest.open("GET", feedUrl, true);
	httpRequest.overrideMimeType('text/plain; charset=x-user-defined');
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
	httpRequest.setRequestHeader("GData-Version", "3.0");
	//httpRequest.setRequestHeader("If-Match", "*");
	var fast_sync = nsPreferences.getBoolPref("gmcont.photo.fast_sync", true);
	if (fast_sync && oldEtag) httpRequest.setRequestHeader("If-None-Match", oldEtag);
	//else httpRequest.setRequestHeader("If-Match", "*");
	
	var me = this;
  var recvFunc = function() {
   	gGCLogger.dump("[Retrieve Photo] "+httpRequest.status+": "+httpRequest.statusText+"\n");
		me.outputMessageToStatusBar("[Retrieve Photo] " + httpRequest.status+": "+httpRequest.statusText);
   	switch (httpRequest.status) {
	    case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
				var file = me.backMgr.getDumpDir(me.account);
				var contentType = httpRequest.getResponseHeader("Content-Type").toLowerCase();
				var ext = contentType.split("/").pop();
				if (ext != "jpeg" && ext != "gif" && ext != "png") {
					ext = null;
					gGCLogger.dump(contentType+" is not supported.\n");
				}
				if (file && ext) {
					file.append("Photos");
					var name = "GoogleContactsPhoto_"+contact.getId().toString().split("/").pop();
					file.append(name+"."+ext);
					if (file.exists()) file.remove(true);
					file.create(file.NORMAL_FILE_TYPE, 0666);
					var bintext = httpRequest.responseText;
					/*
					var maskedChars = [];
  	      for (var i = 0; i < bintext.length; i++) {
    	        maskedChars[i] = String.fromCharCode(bintext.charCodeAt(i) & 0xff);
      	  }
        	bintext = maskedChars.join("");
					*/
          var outstream = Components.classes["@mozilla.org/network/safe-file-output-stream;1"].createInstance(Components.interfaces.nsIFileOutputStream);
					outstream.init(file, 0x02 | 0x08 | 0x20, 0664, 0);
					//outstream.write(httpRequest.responseText, httpRequest.responseText.length);
					outstream.write(bintext, bintext.length);
					if (outstream instanceof Components.interfaces.nsISafeOutputStream) {
						outstream.finish();
					} else {
						outstream.close();
					}

					//the following modify the card and fire the event
					//but it is ignored since the change does not affect contact.isChanged()
			    var ioService = Components.classes['@mozilla.org/network/io-service;1']
                                .getService(Components.interfaces.nsIIOService);
          var photoUri = ioService.newFileURI(file).spec;
					var card = contact.setPhoto(photoUri);
					if (newEtag != oldEtag) {
						/* set new etag to card. but if a photo is changed on server side
						 * after retrieving contacts, it is not latest etag.
						 */ 
						card.setProperty("GmPhotoEtag", newEtag);
					}
					//var addrBookDir = me.getAddressBook();
  				addrBookDir.modifyCard(card);
				}
       	break;
	    case 304: /* Not Modified */
	    	break;
	    case 401: /* Authorization required. */
	    case 403: /* Unsupported standard parameter, or authentication or
	                 Authorization failed. */
 	    	break;
	    case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
				var card = contact.resetPhoto();
				card.deleteProperty("GmPhotoEtag");
				addrBookDir.modifyCard(card);
				break;
			case 500: /* Internal Server Error */
	    default:
	    	break;
  	}
  	contact.clearPhotoManager();
		if (item instanceof Array) {
			me.retrievePhoto(item);
		} else {
			//me.manager.startListening(addrBookURI);
			me.retrievePhoto(null);
		}
 	};

 	this.sendHttpRequest(httpRequest, null, recvFunc);
 	return;
}

GCContacts.prototype.uploadPhoto = function(item, imageURI)
{
	if (!this.gDataAuthToken) {
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		this.manager.notifySyncDone();
		return;
	}
	
	var addrBookURI = this.getAddressBookURI();
	if (this.manager.listening(addrBookURI)) this.manager.stopListening(addrBookURI);
	
	var card = item && item.length ? item.pop() : item;
	if (!card) {
		this.manager.startListening(addrBookURI);
		this.outputMessageToStatusBar("");
		return;
	}

	var photoType = card.getProperty("PhotoType", "")
	if (!imageURI && photoType == "file") {
		imageURI = card.getProperty("PhotoURI", null);
	}
	var id = card.getProperty("GmContactId", null);
	//card has id but not have related contact when card is copied from another Google account
	var contact = id ? this.contacts[id] : null;
	if (!contact || !imageURI) {
		if (item.length) this.uploadPhoto(item, null);
		else this.uploadPhoto(null, null);
		return;
	}
	
	//contact.setPhotoManager(this.manager.photoManager);
	
	var ios = Components.classes['@mozilla.org/network/io-service;1'].getService(Components.interfaces["nsIIOService"]);
	var	URI = ios.newURI(imageURI, null, null);	
	var file = URI.QueryInterface(Components.interfaces.nsIFileURL).file;
	// Make a stream from a file.  
 	var stream = Components.classes["@mozilla.org/network/file-input-stream;1"]  
  	                      .createInstance(Components.interfaces.nsIFileInputStream);  
 	stream.init(file, 0x04 | 0x08, 0644, 0x04); // file is an nsIFile instance     

	var feedUrl = contact.getPhotoLink(true);
	// Try to determine the MIME type of the file  
 	var mimeType = null;
 	try {  
  	var mimeService = Components.classes["@mozilla.org/mime;1"]  
    	       .getService(Components.interfaces.nsIMIMEService);  
   	mimeType = mimeService.getTypeFromFile(file); // file is an nsIFile instance  
 	} catch(e) { 
 		gGCLogger.dump(e+"\n");
		if (item.length) this.uploadPhoto(item, null);
		else this.uploadPhoto(null, null);
 		return;
 	}
 	if (!mimeType) return;
	var ext = mimeType.split("/").pop();
	if (ext != "jpeg" && ext != "gif" && ext != "png") {
		ext = null;
		gGCLogger.dump(mimeType+" is not supported.\n");
		if (item.length) this.uploadPhoto(item, null);
		else this.uploadPhoto(null, null);
		return;
	}
  var httpRequest = new XMLHttpRequest();
  if (nsPreferences.getBoolPref("gmcont.useHTTPMethodOverride",true)) {
 		httpRequest.open("POST", feedUrl, true);
		httpRequest.setRequestHeader("X-HTTP-Method-Override", "PUT");
	} else {
		httpRequest.open("PUT", feedUrl, true);
	}
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
	httpRequest.setRequestHeader("GData-Version", "3.0");
	httpRequest.setRequestHeader("If-Match", "*");
	httpRequest.setRequestHeader("Content-Type", mimeType);
	
	var me = this;
  var recvFunc = function() {
   	gGCLogger.dump("[Upload Photo] "+httpRequest.status+": "+httpRequest.statusText+"\n");
		me.outputMessageToStatusBar("[Upload Photo] " + httpRequest.status+": "+httpRequest.statusText);
   	switch (httpRequest.status) {
	    case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
      	me.outputMessageToStatusBar("The photo is uploaded.");
 				me.retrieve(contact.getId());
      	break;
	    case 304: /* Not Modified */

	    case 401: /* Authorization required. */
	    case 403: /* Unsupported standard parameter, or authentication or
	                 Authorization failed. */
	    case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
	    default:
	    	me.outputMessageToStatusBar("Failed to upload the photo.");
	    	break;
  	}
		if (item.length) me.uploadPhoto(item, null);
		else me.uploadPhoto(null, null);
 	};

 	this.sendHttpRequest(httpRequest, stream, recvFunc);
 	return;
}

GCContacts.prototype.removePhoto = function(item)
{
	if (!this.gDataAuthToken) {
		//alert(this.localeBundle.getString("gmcont.auth_needed"));
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		return;
	}
	
	var addrBookURI = this.getAddressBookURI();
	if (this.manager.listening(addrBookURI)) this.manager.stopListening(addrBookURI);
	
	var abCard = item && item.length ? item.shift() : item;
	if (!abCard) {
		this.manager.startListening(addrBookURI);
		this.outputMessageToStatusBar("");
		return;
	}
	
	//var addr = abCard.primaryEmail;
	var id = abCard.getProperty("GmContactId", null);
	//card has id but not have related contact when card is copied from another Google account
	var contact = id ? this.contacts[id] : null;
	if (!contact) {
		gGCLogger.dump("no contact\n");
		if (item.length) this.removePhoto(item);
		else this.removePhoto(null);
		return;
	}

	contact.setPhotoManager(this.manager.photoManager);
	
	var feedUrl = contact.getPhotoLink();
	if (!feedUrl) {
		gGCLogger.dump("no photo is found in a contact\n");
		if (item.length) this.removePhoto(item);
		else this.removePhoto(null);
		contact.clearPhotoManager();
		return;
	}
	var httpRequest = new XMLHttpRequest();
	if (nsPreferences.getBoolPref("gmcont.useHTTPMethodOverride",true)) {
		httpRequest.open("POST", feedUrl, true);
		httpRequest.setRequestHeader("X-HTTP-Method-Override", "DELETE");
		httpRequest.setRequestHeader("Content-Length", 0);
	} else {
		httpRequest.open("DELETE", feedUrl, true);
	}
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
	//httpRequest.setRequestHeader("GData-Version", "2");
	httpRequest.setRequestHeader("GData-Version", "3.0");
	httpRequest.setRequestHeader("If-Match", "*");
		
	var me = this;
	var recvFunc = function() {
		gGCLogger.dump("[Delete] "+httpRequest.status+": "+httpRequest.statusText+"\n");
 		switch (httpRequest.status) {
    	case 200: /* No error. */
    	case 201: /* Deletion was successful. */
    		//var msg = me.localeBundle.getString("gmcont.delete_succeeded");
    		//me.outputMessageToStatusBar(msg);
    		me.outputMessageToStatusBar("The photo is deleted.");
				abCard = contact.resetPhoto();
				abCard.deleteProperty("GmPhotoEtag");
				var addrBookDir = me.getAddressBook();
				addrBookDir.modifyCard(abCard);
 				me.retrieve(contact.getId());
	      break;
  	  case 401: /* Authorization required. */
    	case 403: /* Unsupported standard parameter, or authentication or
      	           Authorization failed. */
    	case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
				//var msg = me.localeBundle.getString("gmcont.delete_failed");
				//me.outputMessageToStatusBar(msg);
    		//break;
    	default:
    		//var msg = me.localeBundle.getString("gmcont.delete_failed");
      	//me.outputMessageToStatusBar(msg);
    		me.outputMessageToStatusBar("Failed to delete the photo.");
    		break;
  	}
		contact.clearPhotoManager();
  	if (item.length) me.removePhoto(item);
  	else me.removePhoto(null);
	};

 	this.sendHttpRequest(httpRequest, null, recvFunc);
	
	return;
}

GCContacts.prototype.sendHttpRequest = function(httpRequest, body, recvFunc)
{
	try {
		var me = this;
		var wrapperFunc = function() {
			//me.queueNum--;
			//if (me.queueNum == 0) me.manager.progressEnd();
			if (recvFunc) recvFunc.apply();
			me.sendHttpRequestMain(); //send next request
			httpRequest.onload = null; //will avoid memory leak?
		};

		httpRequest.onload = wrapperFunc;
  	var errFunc = function(e) {
  		var msg = me.localeBundle.getString("gmcont.net_err");
  		if (!e.status && !e.statusText) {
  			//alert("Cannot establish connection. Confirm network settings.");
  			this.manager.outputMessageToStatusBar("Cannot establish connection. Confirm network settings.");
  		} else {
  			this.manager.outputMessageToStatusBar(msg+" "+e.status+": "+e.statusText);
  		}
			//me.queueNum--;
			//if (me.queueNum == 0) me.manager.progressEnd();
  		me.sendHttpRequestMain(); //send next request
  		httpRequest.onerror = null; //will avoid memory leak?
  	}
  	httpRequest.onerror = errFunc;
  	if (nsPreferences.getBoolPref("gmcont.bypass_cache", true)) {
  		httpRequest.channel.loadFlags |= Components.interfaces.nsIRequest.LOAD_BYPASS_CACHE;
  	}
    //httpRequest.send(body);
  	this.sendHttpRequestMain(httpRequest, body)
    //if (this.queueNum == 0) this.manager.progressStart();
    //this.queueNum++;
  } catch(e) {
  	this.outputMessageToStatusBar("Exception in requesting");
	  gGCLogger.dump("Exception in requesting: "+e+"\n");
  }
}

GCContacts.prototype.sendHttpRequestMain = function(httpRequest, body)
{
	var preLength = this.requestQueue.length;
	if (httpRequest) this.requestQueue.push([httpRequest, body]);
	
	var nextRequest = this.requestQueue.shift();
	if (nextRequest) {
		nextRequest[0].send(nextRequest[1]);
		if (preLength == 0) this.manager.progressStart();
	} else {
		this.manager.progressEnd();
	}
}

GCContacts.prototype.getLinkHref = function(links, rel)
{
	var ret = null;
	for (var i=0; i<links.length; i++) {
		if (links[i].getAttribute("rel") == rel) {
			ret = links[i].getAttribute("href");
			break;
		}
	}
	
	return ret;
}

GCContacts.prototype.parseFeedXML = function(feed, isGroup)
{
	gGCLogger.dump("[New parser] parse feed\n");


	if (this.busy) {
		gGCLogger.dump("Busy now. Abort sync\n");
		return;
	}
	
	//for dubugging
	try {
		if (nsPreferences.getBoolPref("gmcont.dump_raw_xml", false)) {
			gGCLogger.dump("Dump raw XML\n");
			this.backMgr.dumpRawXMLString(this.account, feed, isGroup, this.startIndex - this.resultsPerPage);
		} else {
			gGCLogger.dump("Skip raw XML dump\n");
		}
	} catch(e) {
		gGCLogger.dump(e);
	}
	
	var links = feed.getElementsByTagNameNS(this.atom, "link");
	this.nextLink = this.getLinkHref(links, "next");

	gGCLogger.dump("Next Link: " + this.nextLink + "\n");
	var entries = feed.getElementsByTagNameNS(this.atom, "entry");
	
	if (isGroup) {
		this.batchMgr.setGroupBatchLink(
			this.getLinkHref(links, 'http://schemas.google.com/g/2005#batch')
		);
		
		var isGoogleApps = !(this.account.match(/@(gmail\.com|googlemail\.com)$/i));
		if (isGoogleApps) gGCLogger.dump("Google Apps account. Skip system groups\n");
		
		var len = entries.length;
		for (var i = 0; i < len; i++) {
			var entry = entries[i];
			var group = this.orgGroups[GCXMLEntry.getIdForEntry(entry)];
			if (group) {
				group.init(entry);
				group.initSystemGroups();
			} else {
				group = new GCGroupEntry(entry);
			}
			if (group.getSystemGroupId() == "Contacts") this.myContactsGroupId = group.getId();
			if (isGoogleApps && group.isSystemGroup() && group.getSystemGroupId() != "Contacts") continue;
			if (this.syncGroupName == "") {
				this.groups[group.getId()] = group;
			} else if (this.syncGroupName == group.getSystemGroupId() || this.syncGroupName == group.getSystemName()) {
				this.syncGroup = group;
			}

			if (this.needYield()) yield;
		}
		gGCLogger.dump("Num entries: "+len+"\n");
		if (this.nextLink) {
			this.retrieveGroups();
			return;
		} else {
			var msg = this.localeBundle.getString("gmcont.reload_done");
			msg = msg.replace(/\%S/, this.account);
			this.outputMessageToStatusBar(msg);
			this.startIndex = 1;
			this.retrieve();
			return;
		}
	} else {
		//snapshot
		this.batchMgr.setContactBatchLink(
			this.getLinkHref(links, 'http://schemas.google.com/g/2005#batch')
		);
		var len = entries.length;
		for (var i=0; i<len; i++) {
			var entry = entries[i];
			var contact = this.orgContacts[GCXMLEntry.getIdForEntry(entry)];
			if (contact) {
				contact.init(entry);
			} else {
				contact = new GCContactEntry(entry);
			}
			this.contacts[contact.getId()] = contact;
			if (this.needYield()) yield;
		}
		gGCLogger.dump("Num entries: "+len+"\n");
		if (this.nextLink) {
			this.retrieve();
			return;
		}
		this.busy = true;
		this.startIndex = 1;
		this.batchMgr.init();
		var me = this;
		var refreshFunc = function() {
			var prevSyncGroupName = nsPreferences.copyUnicharPref("gmcont."+me.account+".prev_sync_group", "");
			if (prevSyncGroupName != me.syncGroupName) {
				gGCLogger.dump("Sync group was changed. Sync as GC is golden.\n");
				nsPreferences.setUnicharPref("gmcont."+me.account+".prev_sync_group", me.syncGroupName);
				me.refreshABCards(true);
			} else {
				me.refreshABCards(false);	
			}
		}
		
		var msg = this.localeBundle.getString("gmcont.reload_done");
		msg = msg.replace(/\%S/, this.account);
		this.outputMessageToStatusBar(msg);
		setTimeout(refreshFunc, 1000);
		nsPreferences.setUnicharPref("gmcont.last_sync_time", this.manager.toRFC3339());
		this.orgContacts = null;
		this.orgGroups = null;
 }
}



GCContacts.prototype.takeSnapShotForTbAb = function()
{
	try {
		gGCLogger.dump("Dump TB addressbook\n");
		var addrBookDir = this.getAddressBook();
		var abURI = this.getAddressBookURI();
		addrBookDir.QueryInterface(Components.interfaces.nsIAbMDBDirectory);
		var db = addrBookDir.database;
		this.backMgr.copyTBAddressBook(this.account, db.dbPath);
	} catch(e) {
		gGCLogger.dump(e);
	}
}

GCContacts.prototype.updateLastSyncDate = function(date)
{
	var msec = date ? date.getTime() : (new Date()).getTime();
	nsPreferences.setIntPref("gmcont."+this.account+".last_sync", parseInt(msec/1000));
}

GCContacts.prototype.getLastSyncDate = function()
{
	//return nsPreferences.getIntPref("gmcont."+this.account+".last_sync", Infinity);
	return nsPreferences.getIntPref("gmcont."+this.account+".last_sync", 0);
}

GCContacts.prototype.getContactByPrimaryAddressOrDispName = function(addr, name)
{
	//if (!addr && !name) return null;
	if (addr) {
		var key;
		for (key in this.contacts) {
			var contact = this.contacts[key];
			if (contact && contact.getPrimaryAddress() == addr) return contact;
		}
	}
	
	if (name) {
		var key;
		for (key in this.contacts) {
			var contact = this.contacts[key];
			if (contact && contact.getDispName() == name) return contact;
		}
	}
	
	return null;
}

GCContacts.prototype.getContactByABCard = function(card)
{
	if (!card) return null;

	var key;
	for (key in this.contacts) {
		var contact = this.contacts[key];
		if (!contact) continue;
		var contCard = contact.getABCard();
		if (contCard && card.equals(contCard)) return contact;
	}

	return null;
}

GCContacts.prototype.getGroupByName = function(name)
{
	if (!name) return null;
	var name1 = name.replace(/:\d+$/,"");
	var key;
	for (key in this.groups) {
		var group = this.groups[key];
		var name2 = group.getTitle().replace(/:\d+$/,"");
		//if (name == group.getTitle()) return group;
		if (name1 == name2) return group;
	}
	
	return null;
}

GCContacts.prototype.getGroupByMailListName = function(name)
{
	if (!name) return null;
	var key;	
	for (key in this.groups) {
		var group = this.groups[key];
		var name1 = name.replace(/:\d+$/,"");
		var name2 = group.getMailList().dirName;
		if (!name2) continue;
		name2 = name2.replace(/:\d+$/,"");
		if (name1 == name2) {
			return group;
		}
		//if (name == group.getMailList().dirName) return group;
	}
	
	return null;
}

GCContacts.prototype.getGroupByMailList = function(mailList)
{
	if (!mailList) return null;
	var list1 = mailList.QueryInterface(Components.interfaces.nsIAbDirectory);
	//var ret = this.getGroupByMailListName(list1.dirName);
	var ret = null;
	var id = nsPreferences.copyUnicharPref("gmcont."+this.account+"."+list1.URI+".groupId", null);
	if (id) {
		ret = this.groups[id];
	}
	if (!ret) {
		var key;	
		for (key in this.groups) {
			var group = this.groups[key];
			var URI1 = list1.URI;
			var list2 = group.getMailList();
			if (!list2) continue;
			try {
				var URI2 = list2.URI;
				if (URI1 == URI2) {
					ret = group;
					break;
				}
			} catch(e){ //list2 is being created
			}
		}
	}
	if (!ret) ret = this.getGroupByMailListName(list1.dirName);
	
	return ret;
}

GCContacts.prototype.removeGroup = function(mailList)
{
	if (!this.gDataAuthToken) {
		//alert(this.localeBundle.getString("gmcont.auth_needed"));
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		return;
	}
	mailList.QueryInterface(Components.interfaces.nsIAbDirectory);
	var group = this.getGroupByMailList(mailList);
	if (group) {
		if (group.isSystemGroup()) {
			gGCLogger.dump("Cannot delete system group\n");
			return;
		}
		var feedUrl = group.getEditLink();
		var httpRequest = new XMLHttpRequest();
		if (nsPreferences.getBoolPref("gmcont.useHTTPMethodOverride",true)) {
			httpRequest.open("POST", feedUrl, true);
			httpRequest.setRequestHeader("X-HTTP-Method-Override", "DELETE");
			httpRequest.setRequestHeader("Content-Length", 0);
		} else {
			httpRequest.open("DELETE", feedUrl, true);
		}
		httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
		//httpRequest.setRequestHeader("GData-Version", "2");
		httpRequest.setRequestHeader("GData-Version", "3.0");
		httpRequest.setRequestHeader("If-Match", "*");

		var me = this;
  	var recvFunc = function() {
			gGCLogger.dump("[Delete Group] "+httpRequest.status+": "+httpRequest.statusText+"\n");
   		switch (httpRequest.status) {
	    	case 200: /* No error. */
      	case 201: /* Deletion was successful. */
      		//me.groups[group.getId()] = null;
      		//delete me.groups[group.getId()];
      		var msg = me.localeBundle.getString("gmcont.delete_succeeded");
      		me.outputMessageToStatusBar(msg);
      		//me.retrieveGroups();
  	      break;
    	  case 401: /* Authorization required. */
      	case 403: /* Unsupported standard parameter, or authentication or
        	           Authorization failed. */
      	case 409: /* Conflict */
				case 400: /* Bad Request */
				case 404: /* Not Found */
				case 500: /* Internal Server Error */
					var msg = me.localeBundle.getString("gmcont.delete_failed");
					me.outputMessageToStatusBar(msg);
      		break;
      	default:
      		var msg = me.localeBundle.getString("gmcont.delete_failed");
	      	me.outputMessageToStatusBar(msg);
      		break;
    	}
 		};

	 	this.sendHttpRequest(httpRequest, null, recvFunc);
		//this.contacts[group.getId()] = null;
 		this.deleteGroupEntry(group);
	}
	
	return;
}

GCContacts.prototype.removeContact = function(abCard)
{
	if (!this.gDataAuthToken) {
		//alert(this.localeBundle.getString("gmcont.auth_needed"));
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		return;
	}
	abCard.QueryInterface(Components.interfaces.nsIAbCard);
	//var addr = abCard.primaryEmail;
	//var contact = this.getContactByABCard(abCard);
	var id = abCard.getProperty("GmContactId", null);
	//card has id but not have related contact when card is copied from another Google account
	var contact = id ? this.contacts[id] : null;
	if (contact) {
		var delMapMode = nsPreferences.getIntPref("gmcont.map_del_mode", 0);
		if (!this.syncGroup) delMapMode = null;
	 	if (delMapMode == 2) delMapMode = this.askDelOperationForGroupMapping();
   	if (delMapMode == 1) {
   		gGCLogger.dump("Delete from the group\n");
   		this.removeContactFromMappedGroup(contact);
   		return;
   	}
   	gGCLogger.dump("Delete from Google\n");
		var feedUrl = contact.getEditLink();
		var httpRequest = new XMLHttpRequest();
		if (nsPreferences.getBoolPref("gmcont.useHTTPMethodOverride",true)) {
			httpRequest.open("POST", feedUrl, true);
			httpRequest.setRequestHeader("X-HTTP-Method-Override", "DELETE");
			httpRequest.setRequestHeader("Content-Length", 0);
		} else {
			httpRequest.open("DELETE", feedUrl, true);
		}
		httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
		//httpRequest.setRequestHeader("GData-Version", "2");
		httpRequest.setRequestHeader("GData-Version", "3.0");
		httpRequest.setRequestHeader("If-Match", "*");
		
		var me = this;
  	var recvFunc = function() {
			gGCLogger.dump("[Delete] "+httpRequest.status+": "+httpRequest.statusText+"\n");
   		switch (httpRequest.status) {
	    	case 200: /* No error. */
      	case 201: /* Deletion was successful. */
      		//me.contacts[contact.getId()] = null;
      		//delete me.contacts[contact.getId()];
      		var msg = me.localeBundle.getString("gmcont.delete_succeeded");
      		me.outputMessageToStatusBar(msg);
  	      break;
    	  case 401: /* Authorization required. */
      	case 403: /* Unsupported standard parameter, or authentication or
        	           Authorization failed. */
      	case 409: /* Conflict */
				case 400: /* Bad Request */
				case 404: /* Not Found */
				case 500: /* Internal Server Error */
					var msg = me.localeBundle.getString("gmcont.delete_failed");
					me.outputMessageToStatusBar(msg);
      		break;
      	default:
      		var msg = me.localeBundle.getString("gmcont.delete_failed");
	      	me.outputMessageToStatusBar(msg);
      		break;
    	}
 		};

	 	this.sendHttpRequest(httpRequest, null, recvFunc);
		//this.contacts[contact.getId()] = null;
	 	//delete this.contacts[contact.getId()];
 		this.deleteContactEntry(contact);
	}
	
	return;
}

GCContacts.prototype.removeContactFromMappedGroup = function(contact)
{
	contact.unbelongToGroup(this.syncGroup);
	var feedUrl = contact.getEditLink();
	var method = "PUT";
	this.manager.startListening(this.getAddressBookURI());
	//now update GContact
  var httpRequest = new XMLHttpRequest();
  if (method == "PUT" && nsPreferences.getBoolPref("gmcont.useHTTPMethodOverride",true)) {
 		httpRequest.open("POST", feedUrl, true);
		httpRequest.setRequestHeader("X-HTTP-Method-Override", "PUT");
	} else {
		httpRequest.open(method, feedUrl, true);
	}
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
	//httpRequest.setRequestHeader("GData-Version", "2");
	httpRequest.setRequestHeader("GData-Version", "3.0");
	httpRequest.setRequestHeader("If-Match", "*");
	
	var me = this;
  var recvFunc = function() {
  	gGCLogger.dump("[Update] "+httpRequest.status+": "+httpRequest.statusText+"\n");
   	switch (httpRequest.status) {
	    case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
      var feed = httpRequest.responseXML;
 			//var entryXML = new XML(feed.replace(/\<\?xml.+\?\>/,""));
				contact.init(feed.documentElement);
				var msg = me.localeBundle.getString("gmcont.update_succeeded");
   			me.outputMessageToStatusBar(msg);
     		//me.retrieve(id);
        break;
      case 401: /* Authorization required. */
      case 403: /* Unsupported standard parameter, or authentication or
                   Authorization failed. */
      case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
				var msg = me.localeBundle.getString("gmcont.update_failed")
				me.outputMessageToStatusBar(msg);
      	break;
      default:
 				var msg = me.localeBundle.getString("gmcont.update_failed")
				me.outputMessageToStatusBar(msg);
      	break;
    }
  };

 	var postBody = contact.toXMLString(true);
  //var postBody = contact.getEntry();
  var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
	                  createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
 	converter.charset = "UTF-8";
 	var stream = converter.convertToInputStream(postBody);
 	httpRequest.setRequestHeader("Content-Type", "application/atom+xml");
 	
	this.sendHttpRequest(httpRequest, stream, recvFunc);
}

GCContacts.prototype.updateGroup = function(mailList, skipListUpdate)
{
	mailList.QueryInterface(Components.interfaces.nsIAbDirectory);
	var addrBookDir = this.getAddressBook();
	if (!addrBookDir.hasDirectory(mailList)) return;
	
	if (!this.gDataAuthToken) {
		//alert(this.localeBundle.getString("gmcont.auth_needed"));
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		return;
	}

	var group = this.getGroupByMailList(mailList);

	var feedUrl = "";
	var method = "";
	var newGroup = false;
	
	if (group) {
		gGCLogger.dump(group.getTitle()+"\n");
		if (!group.isChanged(mailList)) {
			gGCLogger.dump("group modify not needed\n");
			return;
		}
		gGCLogger.dump("modify group\n");
		//group.setMailList(mailList, this.account);
		if (group.isSystemGroup() || !group.isChanged(mailList, true)) {
			gGCLogger.dump("System group or group name is not changed\n");
			this.updateGroupForContact(group);
			var msg = this.localeBundle.getString("gmcont.update_succeeded");
 			this.outputMessageToStatusBar(msg);
			return;
		}
		feedUrl = group.getEditLink();
		method = "PUT";
	}	else {
		gGCLogger.dump("add group\n");
		group = new GCGroupEntry(); //create contact temporary
		group.setMailList(mailList, this.account);
		feedUrl = this.baseGroupsFeedUrl;
		method = "POST";
		newGroup = true;
	}
	group.updateGGroupFromMailList(mailList);
	this.manager.startListening(this.getAddressBookURI());
	//now update GContact
  var httpRequest = new XMLHttpRequest();
  if (method == "PUT" && nsPreferences.getBoolPref("gmcont.useHTTPMethodOverride",true)) {
 		httpRequest.open("POST", feedUrl, true);
		httpRequest.setRequestHeader("X-HTTP-Method-Override", "PUT");
	} else {
		httpRequest.open(method, feedUrl, true);
	}
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
	//httpRequest.setRequestHeader("GData-Version", "2");
	httpRequest.setRequestHeader("GData-Version", "3.0");
	if (!newGroup) httpRequest.setRequestHeader("If-Match", "*");
	
	var me = this;
  var recvFunc = function() {
  	gGCLogger.dump("[Update Group] "+httpRequest.status+": "+httpRequest.statusText+"\n");
   	switch (httpRequest.status) {
	    case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
      	var feed = httpRequest.responseXML;
 				//var entryXML = new XML(feed.replace(/\<\?xml.+\?\>/,""));
	    	if (newGroup) {
					var abURI = me.getAddressBookURI();
     			me.manager.stopListening(abURI);
     			try {
						var grp = new GCGroupEntry(feed.documentElement);
						var grpId = grp.getId();
						me.groups[grpId] = grp;
						grp.setMailList(mailList, me.account);
						var updateContactsFunc = function() {
	     				me.manager.startListening(abURI);
							if (!skipListUpdate) me.updateGroupForContact(grp);
     					me.manager.startListening(abURI);
		     			var msg = me.localeBundle.getString("gmcont.add_succeeded");
		     			me.outputMessageToStatusBar(msg);
						};
						setTimeout(updateContactsFunc, 1000);
     			} catch(e) {
     				me.manager.startListening(abURI);
	     			var msg = me.localeBundle.getString("gmcont.add_succeeded");
	     			me.outputMessageToStatusBar(msg);
     			}

	     	} else {
					group.init(feed.documentElement);
					group.setMailList(mailList, me.account);
					var updateContactsFunc = function() {
						if (!skipListUpdate) me.updateGroupForContact(group);
						var msg = me.localeBundle.getString("gmcont.update_succeeded");
     				me.outputMessageToStatusBar(msg);
					};
					setTimeout(updateContactsFunc, 1000);
	     		//me.retrieve(id);
	     	}
        break;
      case 401: /* Authorization required. */
      case 403: /* Unsupported standard parameter, or authentication or
                   Authorization failed. */
      case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
				var msg = newGroup 
								? me.localeBundle.getString("gmcont.add_failed")
								: me.localeBundle.getString("gmcont.update_failed")
				me.outputMessageToStatusBar(msg);
      	break;
      default:
 				var msg = newGroup 
								? me.localeBundle.getString("gmcont.add_failed")
								: me.localeBundle.getString("gmcont.update_failed")
				me.outputMessageToStatusBar(msg);
      	break;
    }
  };

 	var postBody = group.toXMLString(true);
  //var postBody = group.getEntry();
  var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
	                  createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
 	converter.charset = "UTF-8";
 	var stream = converter.convertToInputStream(postBody);
 	httpRequest.setRequestHeader("Content-Type", "application/atom+xml");
 	
	this.sendHttpRequest(httpRequest, stream, recvFunc);
}

GCContacts.prototype.getContactIdsForGroup = function(grp)
{
	var id = grp.getId();
	var ret = new Array();
	var key;
	for (key in this.contacts) {
		var contact = this.contacts[key];
		var grps = contact.getBelongingGroupIds();
		for (var i=0; i<grps.length; i++) {
			if (grps[i] == id) {
				//ret.push(key);
				ret[key] = 1;
				break;
			}
		}
	}

	return ret;
}

GCContacts.prototype.updateGroupForContact = function(grp, batch)
{
	var orgContactKeys = this.getContactIdsForGroup(grp); //key => 1 hash
	var updated = new Array();
	var list = grp.getMailList();
	if (!list || !list.addressLists) {
		gGCLogger.dump("list is not set\n");
		return updated;
	}
	
	var total = list.addressLists.length;
  for ( var i = 0;  i < total; i++ ) {
  	var card = list.addressLists.queryElementAt(i, Components.interfaces.nsIAbCard);
  	var contactId = card.getProperty("GmContactId", null);
  	var contact = contactId ? this.contacts[contactId] : null;
		if (contact) {
			if (orgContactKeys[contactId]) {
				//still in the list. need not to update
				delete orgContactKeys[contactId];
			} else {
				//newly added to the list
				contact.belongToGroup(grp);
				if (!batch) this.updateContact(contact.getABCard(), true);
				updated.push(contact);
			}
		} else if (!batch) { //card added to an address book and a list at the same time with intr. mode
			this.updateContact(card, true, grp);
		}
  }
  
  //removed from the list
  for (var key in orgContactKeys) {
		var contact = this.contacts[key];
		if (!contact) continue;
		contact.unbelongToGroup(grp);
		if (!batch) this.updateContact(contact.getABCard(), true);
		updated.push(contact);
  }

  return updated;
}

GCContacts.prototype.updateContact = function(abCard, force, belongedGroup)
{
	var addrBookDir = this.getAddressBook();
	if (!abCard || !addrBookDir.hasCard(abCard)) return;
	
	if (!this.gDataAuthToken) {
		//alert(this.localeBundle.getString("gmcont.auth_needed"));
		this.outputMessageToStatusBar(this.localeBundle.getString("gmcont.auth_needed"));
		return;
	}
  
	//abCard.QueryInterface(Components.interfaces.nsIAbMDBCard);
	abCard.QueryInterface(Components.interfaces.nsIAbCard);
	//var id = abCard.getStringAttribute("GmContactId");
	var id = abCard.getProperty("GmContactId", null);
	//card has id but not have related contact when card is copied from another Google account
	var contact = id ? this.contacts[id] : null;
	var feedUrl = "";
	var method = "";
	var newContact = false;
	if (contact) {
		//contact = this.contacts[id];
		abCard.QueryInterface(Components.interfaces.nsIAbCard);
		if (!force && !contact.isChanged(abCard)) {
			gGCLogger.dump("modify not needed\n");
			return;
		}
		gGCLogger.dump("modify\n");
		feedUrl = contact.getEditLink();
		method = "PUT";
		try {
			var needUp = contact.updateGContactFromCard(abCard);
			//if (!contact.updateGContactFromCard(abCard) && !force) return;
			if (!needUp && !force) return;
		}catch(e){
			gGCLogger.dump("Error in uploading contact\n");
			gGCLogger.dump(e+"\n");
		}
	}	else {
		contact = new GCContactEntry(); //create contact temporary
		abCard.QueryInterface(Components.interfaces.nsIAbCard);
		contact.setABCard(abCard);
		contact.updateGContactFromCard(abCard)
		if (this.syncGroup) {
			contact.belongToGroup(this.syncGroup);
		} else if (nsPreferences.getBoolPref("gmcont.add_to_mycontacts", false) && this.myContactsGroupId) {
			contact.belongToGroup(this.myContactsGroupId);
		}
		if (belongedGroup) contact.belongToGroup(belongedGroup);
		feedUrl = this.baseFeedUrl;
		method = "POST";
		newContact = true;
	}
	
	if (abCard.getProperty("GmContactEtag", null)) {
		this.manager.startListening(this.getAddressBookURI());
		abCard.deleteProperty("GmContactEtag");
		addrBookDir.modifyCard(abCard);
		this.manager.startListening(this.getAddressBookURI());
	}
	//if (!contact.updateGContactFromCard(abCard)) return;

	//this.manager.startListening(this.getAddressBookURI());
	//now update GContact
  var httpRequest = new XMLHttpRequest();
  if (method == "PUT" && nsPreferences.getBoolPref("gmcont.useHTTPMethodOverride",true)) {
 		httpRequest.open("POST", feedUrl, true);
		httpRequest.setRequestHeader("X-HTTP-Method-Override", "PUT");
	} else {
		httpRequest.open(method, feedUrl, true);
	}
	httpRequest.setRequestHeader("Authorization", "GoogleLogin auth="+this.gDataAuthToken);
	//httpRequest.setRequestHeader("GData-Version", "2");
	httpRequest.setRequestHeader("GData-Version", "3.0");
	if (!newContact) httpRequest.setRequestHeader("If-Match", "*");
	
	var me = this;
  var recvFunc = function() {
  	gGCLogger.dump("[Update] "+httpRequest.status+": "+httpRequest.statusText+"\n");
   	switch (httpRequest.status) {
	    case 200: /* No error. */
      case 201: /* Creation of a resource was successful. */
      var feed = httpRequest.responseXML;
 			//var entryXML = new XML(feed.replace(/\<\?xml.+\?\>/,""));
	     	if (newContact) {
					var abURI = me.getAddressBookURI();
     			me.manager.stopListening(abURI);
     			try {
						var cont = new GCContactEntry(feed.documentElement);
						var contactId = cont.getId();
						me.contacts[contactId] = cont;
						abCard.QueryInterface(Components.interfaces.nsIAbCard);
						abCard.setProperty("GmBookUri", abURI);
						abCard.setProperty("GmContactId", contactId);
						addrBookDir.modifyCard(abCard);
						cont.setABCard(abCard);
    	 			me.manager.startListening(abURI);
     			} catch(e) {
     				gGCLogger.dump("Exception in add card"+e+"\n");
     				me.manager.startListening(abURI);
     			}
     			var msg = me.localeBundle.getString("gmcont.add_succeeded");
     			me.outputMessageToStatusBar(msg);
	     	} else {
					contact.init(feed.documentElement);
					var msg = me.localeBundle.getString("gmcont.update_succeeded");
     			me.outputMessageToStatusBar(msg);
	     		//me.retrieve(id);
	     	}
        break;
      case 401: /* Authorization required. */
      case 403: /* Unsupported standard parameter, or authentication or
                   Authorization failed. */
      case 409: /* Conflict */
			case 400: /* Bad Request */
			case 404: /* Not Found */
			case 500: /* Internal Server Error */
				var msg = newContact 
								? me.localeBundle.getString("gmcont.add_failed")
								: me.localeBundle.getString("gmcont.update_failed")
				me.outputMessageToStatusBar(msg);
				gGCLogger.dump("Method: "+method+"\n");
				gGCLogger.dump("Feed URL: "+feedUrl+"\n");
				gGCLogger.dump("Token: "+me.gDataAuthToken+"\n");
				gGCLogger.dump(contact.toXMLString(true)+"\n");
      	break;
      default:
 				var msg = newContact 
								? me.localeBundle.getString("gmcont.add_failed")
								: me.localeBundle.getString("gmcont.update_failed")
				me.outputMessageToStatusBar(msg);
				gGCLogger.dump(contact.toXMLString(true)+"\n");
      	break;
    }
  };

  var postBody = contact.toXMLString(true);
  var converter = Components.classes["@mozilla.org/intl/scriptableunicodeconverter"].
	                  createInstance(Components.interfaces.nsIScriptableUnicodeConverter);
 	converter.charset = "UTF-8";
 	var stream = converter.convertToInputStream(postBody);
 	httpRequest.setRequestHeader("Content-Type", "application/atom+xml");
	this.sendHttpRequest(httpRequest, stream, recvFunc);
}

GCContacts.prototype.outputMessageToStatusBar = function(str)
{
	this.manager.outputMessageToStatusBar(str);
}

GCContacts.prototype.printContacts = function()
{
	for (var key in this.contacts) {
		var contact = this.contacts[key];
		contact.print();
	}
}

GCContacts.prototype.removeDuplicates = function()
{
	if (!this.gDataAuthToken) {
		//alert(this.localeBundle.getString("gmcont.auth_needed"));
		this.outputMessageToStatusBar("Reload contacts before run this function");
		return;
	}
	
	var promptService = Components.classes["@mozilla.org/embedcomp/prompt-service;1"].getService(Components.interfaces.nsIPromptService);
	var ok = promptService.confirm(window, 
		"Google Contacts Notification",
		"This is experimental function and you cannot undo. Is it OK to start?"
	);
	if (!ok) {
		return;
	}
	
	this.takeSnapShotForTbAb();
	
	//yield
	var gen = this.removeDuplicatesMain();
	var genFunc = function() {
		try {
			gen.next();
			setTimeout(genFunc, 100);
		} catch (e if e instanceof StopIteration) {
		} catch (e) {
			gGCLogger.dump(e+"\n");
		}
	}
	setTimeout(genFunc, 100);
}

GCContacts.prototype.removeDuplicatesMain = function()
{
	gGCLogger.dump("Finding duplicates ...\n");
	this.outputMessageToStatusBar("Removing duplicates ...");
	var addrBookDir = this.getAddressBook();
	var abURI = this.getAddressBookURI();
	this.manager.progressStart();
	this.manager.stopListening(abURI, true);
 	this.batchMgr.init();
 	
	var checkedIds = []; //same ID. Duplicate in TB.
	var checkedKeys = []; //same contacts but ID is different. Duplicate in Google
	var delCardsMutableArray = Components.classes["@mozilla.org/array;1"].createInstance(Components.interfaces.nsIMutableArray);
	var searchQuery = "?(not(IsMailList,=,TRUE))";
  var childCards = this.abManager.getDirectory(abURI + searchQuery).childCards;
  var childCardsArray = gGCUtils.enumToArray(childCards, Components.interfaces.nsIAbCard);
  for (var i=0; i<childCardsArray.length; i++) {
	  var card = childCardsArray[i];
	  var id = card.getProperty("GmContactId", null);
	  var name = card.getProperty("DisplayName", "");
	  var primal = card.getProperty("PrimaryEmail", "");
	  if (primal.indexOf("@pseudo.googlecontacts") >= 0) primal = "";
	  var second = card.getProperty("SecondEmail", "");
	  var key = name + "/" + primal + "/" + second;
	  var contact = id ? this.contacts[id] : null;
	  if (!contact) { //No Id. Invalid card.
	  	delCardsMutableArray.appendElement(card, false);
	  } else if (checkedIds[id]) { //Duplicated in TB
	  	delCardsMutableArray.appendElement(card, false);
	  } else if (checkedKeys[key]) { //Duplicated in Google
	  	checkedIds[id] = true;
	  	this.batchMgr.registerDelEntry(contact);
	  	delCardsMutableArray.appendElement(card, false);
	  } else { //not a duplicate
	  	checkedIds[id] = true;
	  	checkedKeys[key] = true;
	  }
 		if (this.needYield()) yield;
  }
	
  this.recordUndo = false;
  this.gEnd = false;
  this.tEnd = false;
  var me = this;
  this.batchMgr.setCallback(function(){
  		me.recordUndo = true;
  		gGCLogger.dump("Removing at Google side is done\n");
  		me.outputMessageToStatusBar("Removing at Google side is done\n");
  		if (me.tEnd) { //TB side already finished
  			me.tEnd = false;
  			me.gEnd = false;
  			alert("done.");
  		} else {
  			me.gEnd = true; //Say Google side is done
  		}
  	});
	if (!this.batchMgr.doBatch()) {
		gGCLogger.dump("Removing at Google side is done\n");
		me.outputMessageToStatusBar("Removing at Google side is done\n");
		this.gEnd = true; //Say Google side is done
	}
	if (delCardsMutableArray.length > 0) {
		gGCLogger.dump("Remove "+delCardsMutableArray.length+"cards\n");
		addrBookDir.deleteCards(delCardsMutableArray);
	}
	gGCLogger.dump("Removing at Thunderbird side is done\n");
	this.outputMessageToStatusBar("Removing at Thunderbird side is done\n");
	this.manager.startListening(abURI, true);
	this.manager.progressEnd();
	
	if (this.gEnd) { //Google side already finished
		me.tEnd = false;
		me.gEnd = false;
		alert("done.");
	} else {
		this.tEnd = true; //Say TB side is done
	}
}
