//Super class for contact and group
function GCXMLEntry(entry)
{
	//Namespace
	this.atom = "http://www.w3.org/2005/Atom";
	this.openSearch = "http://a9.com/-/spec/opensearch/1.1/";
	this.gContact = "http://schemas.google.com/contact/2008";
	this.gd = "http://schemas.google.com/g/2005";
	this.batch = "http://schemas.google.com/gdata/batch";
	
	this.localeBundle = document.getElementById("GmContBundle");
	this.itemObj = {};
	this.useDiskCache = nsPreferences.getBoolPref("gmcont.use_disk_cache", true);

	this.entry = entry;

	this.cacheFileName = null;
	this.saveCache();
}

GCXMLEntry.getIdForEntry = function(entry)
{
	return entry.getElementsByTagNameNS("http://www.w3.org/2005/Atom", "id")[0].firstChild.nodeValue;
}

GCXMLEntry.prototype.getCacheDir = function()
{
	var backupMgr = new GCBackupManager();
	var file = backupMgr.getCacheDir();
	
	return file;
}

GCXMLEntry.prototype.loadCache = function()
{
	if (!this.useDiskCache || !this.cacheFileName) return false;
	var file = this.getCacheDir();
	if (!file) return false;
	file.append(this.cacheFileName);
	if (!file.exists()) return false;

	var stream = Components.classes['@mozilla.org/network/file-input-stream;1'].createInstance(Components.interfaces.nsIFileInputStream);
  try {
	  stream.init(file, 1, 0, false);
  	var converterStream = Components
				.classes['@mozilla.org/intl/converter-input-stream;1']
				.createInstance(Components.interfaces.nsIConverterInputStream);
		converterStream.init(stream, "UTF-8", stream.available(),
												 Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
		var cacheObj = {};
		converterStream.readString(stream.available(), cacheObj);
		converterStream.close();
		stream.close();
		
		var parser = new DOMParser();
		var dom = parser.parseFromString(cacheObj.value, "text/xml");
		if (dom.documentElement.nodeName == "parsererror") {
			this.entry = null;
			gGCLogger.dump("Failed to load cache\n");
			return false;
		} else {
			this.entry = dom.documentElement;
		}
	} catch(e) {
		gGCLogger.dump(e);
		return false;
	}

	return true;
}

GCXMLEntry.prototype.setCacheFileName = function(id)
{
	this.cacheFileName = id ? id : this.getId();
	this.cacheFileName = this.cacheFileName.replace(/[@:/%]/g,"_");
}

GCXMLEntry.prototype.saveCache = function()
{
	if (!this.entry || !this.useDiskCache) return false;

	var file = this.getCacheDir();
	//this.cacheFileName = this.getId().toString();
	//this.cacheFileName = this.cacheFileName.replace(/[@:/%]/g,"_");
	this.setCacheFileName();
	file.append(this.cacheFileName);
	
	if (file.exists()) file.remove(true);
	file.create(file.NORMAL_FILE_TYPE, 0666);

	var fileStream = Components.classes['@mozilla.org/network/file-output-stream;1']
										.createInstance(Components.interfaces.nsIFileOutputStream);
	fileStream.init(file, 2, 0x200, false);
	
	var converterStream = Components
			.classes['@mozilla.org/intl/converter-output-stream;1']
			.createInstance(Components.interfaces.nsIConverterOutputStream);
			
	try {										 
		converterStream.init(fileStream, "UTF-8", 0,
											 Components.interfaces.nsIConverterInputStream.DEFAULT_REPLACEMENT_CHARACTER);
		converterStream.writeString(this.toXMLString());
	} catch(e) {
		gGCLogger.dump(e);
		return false;
	}

	converterStream.close();
	fileStream.close();
	
	return true;
}

GCXMLEntry.prototype.removeCache = function()
{
	if (!this.cacheFileName) return false;
	
	var file = this.getCacheDir();
	file.append(this.cacheFileName);
	if (file.exists()) file.remove(false);

	return true;
}

GCXMLEntry.prototype.purgeMemCache = function()
{
	if (!this.useDiskCache) return false;
	this.entry = null;
	return true;
}
/*
GCXMLEntry.prototype.toString = function(rmEtag)
{
	if (!this.entry && !this.loadCache()) return "";
	var ret = null;
	var serializer = new XMLSerializer();
	if (rmEtag) {
		var clone = this.entry.cloneNode(true);
		clone.removeAttributeNS(this.gd, "etag");
		ret = serializer.serializeToString(clone);
	} else {
		ret = serializer.serializeToString(this.entry);
	}
	
	return ret;
}
*/
GCXMLEntry.prototype.toXMLString = function(rmEtag)
{
	if (!this.entry && !this.loadCache()) return "";
	var ret = null;
	var serializer = new XMLSerializer();
	if (rmEtag) {
		var clone = this.entry.cloneNode(true);
		clone.removeAttributeNS(this.gd, "etag");
		ret = serializer.serializeToString(clone);
	} else {
		ret = serializer.serializeToString(this.entry);
	}
	
	return ret;
}

GCXMLEntry.prototype.init = function(entryDOM)
{
	this.entry = entryDOM;
	this.saveCache();
}

GCXMLEntry.prototype.getEntry = function()
{
	if (!this.entry && !this.loadCache()) return null;
	return this.entry;
}

GCXMLEntry.prototype.toRFC3339 = function(aDate)
{
	var date = aDate ? aDate : new Date();
	var str = date.getUTCFullYear()                         
					+ "-" + ("0"+(date.getUTCMonth()+1)).substr(-2)
					+ "-" + ("0"+date.getUTCDate()).substr(-2)
					+ "T" + ("0"+date.getUTCHours()).substr(-2)
					+ ":"	+ ("0"+date.getUTCMinutes()).substr(-2)
					+ ":" + ("0"+date.getUTCSeconds()).substr(-2)
					+ "."	+ ("00"+date.getUTCMilliseconds()).substr(-3)
					+ "Z";
					
	return str;
}

GCXMLEntry.prototype.getId = function()
{
	if (!this.entry && !this.loadCache()) return "";
	var ret = null;
	try {
		ret = this.entry.getElementsByTagNameNS(this.atom, "id")[0].firstChild.nodeValue;
	}catch(e){
		ret = null;
	}
	
	return ret;
}

GCXMLEntry.prototype.getTitle = function()
{
	if (!this.entry && !this.loadCache()) return "";
	var ret = null;
	try {
		ret = this.entry.getElementsByTagNameNS(this.atom, "title")[0].firstChild.nodeValue;
	}catch(e){
		ret = null;
	}
	
	return ret;
}

GCXMLEntry.prototype.getLinkHref = function(rel)
{
	if (!this.entry && !this.loadCache()) return null;
	var ret = null;
	var links = this.entry.getElementsByTagNameNS(this.atom, "link");
	for (var i=0; i<links.length; i++) {
		if (links[i].getAttribute("rel") == rel) {
			ret = links[i].getAttribute("href");
			break;
		}
	}
	
	return ret;
}

GCXMLEntry.prototype.getEditLink = function()
{
	return this.getLinkHref("edit");
}

GCXMLEntry.prototype.getSelfLink = function()
{
	return this.getLinkHref("self");
}

GCXMLEntry.prototype.getEntryEtag = function()
{
	if (!this.entry && !this.loadCache()) return null;
	return this.entry.getAttributeNS(this.gd, "etag");
}

GCXMLEntry.prototype.getUpdatedTimeInMilliSec = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var atom = this.atom;
	var updated = this.entry.getElementsByTagNameNS(this.atom, "updated")[0].firstChild.nodeValue;
	var updatedArray = updated.split("T");
	var dateArray = updatedArray[0].split("-");
	var timeArray = updatedArray[1].split(":");

	var secArray = timeArray[2].split(".");
	timeArray[2] = secArray[0].replace(/Z$/,"");
	if (secArray[1]) {		
		timeArray.push(secArray[1].replace(/Z$/,""));
	}
	
	dateArray = dateArray.concat(timeArray);
	//remove prefix 0
	for (var i=0; i<dateArray.length; i++) {
		dateArray[i] = dateArray[i].replace(/^0+/,"");
	}
	
	var ms = Date.UTC(dateArray[0],   //Year
									dateArray[1]-1, //Month
									dateArray[2],   //Day
									dateArray[3],   //Hours
									dateArray[4],   //Minutes
									dateArray[5],   //Seconds
									dateArray[6]);  //Milliseconds
	return ms;
}

GCXMLEntry.prototype.setTBOnlyItemsToGContact = function(abItem)
{
	if (!this.entry && !this.loadCache()) return;
	
	var oldProps = this.entry.getElementsByTagNameNS(this.gd, "extendedProperty");
	for (var i=oldProps.length-1; i>=0; i--) {
		this.entry.removeChild(oldProps[i]);
	}

	var key = "";
	for (key in this.itemObj) {
		var itemTB = this.itemObj[key];
		var itemValueArray = new Array();
		for (var i=0; i<itemTB.length; i++) {
			if (itemTB[i].indexOf("-") == 0) continue; //obsolete item
			var itemValue = abItem instanceof Components.interfaces.nsIAbCard
											? abItem.getProperty(itemTB[i], null)
											: abItem[itemTB[i]];
			if (itemValue == null || itemValue == undefined) itemValue = "";
			itemValue = itemValue.toString().replace(/\//g,'_gcslash_'); //escape
			itemValueArray.push(itemTB[i]+"="+itemValue);
		}
		if (itemValueArray.length > 0) {
			try {
				var xmlPiece = document.createElementNS(this.gd, "extendedProperty");
				xmlPiece.setAttribute("name", key);
				xmlPiece.setAttribute("value", itemValueArray.join("/"));
				this.entry.appendChild(xmlPiece);
			} catch(e) {
				gGCLogger.dump(e+"\n");
			}
		}
	}
}

GCXMLEntry.prototype.loadTBOnlyItemsFromGContact = function(abItem)
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return;
	
	var props = this.entry.getElementsByTagNameNS(this.gd, "extendedProperty");
	for (var i=0; i<props.length; i++) {
		var itemTB = this.itemObj[key];
		if (!itemTB) continue;
		var key = props[i].getAttribute("name");
		var itemValueStr = props[i].getAttribute("value");
		if (itemValueStr) {
			var itemValueArray = itemValueStr.split("/");
			if (itemValueStr.indexOf("=") != -1) { //0.2.60 or above
				for (var i=0; i<itemValueArray.length; i++) {
					var item = itemValueArray[i].replace(/_gcslash_/g,'/'); //unescape _gcslash_
					var itemNameAndValue = item.split("=");
					var itemName = itemNameAndValue.shift();
					var itemValue = itemNameAndValue.join("=");
					itemName = this.convertTB2PropertyToTB3(itemName);
					if (itemValue === "true") itemValue = true;
					else if (itemValue === "false") itemValue = false;
					try {
						if (abItem instanceof Components.interfaces.nsIAbCard) {
							abItem.setProperty(itemName,itemValue);
						} else {
							abItem[itemName] = itemValue;
						}
						//abItem[itemName] = itemValue;
					} catch(e) {
						gGCLogger.dump(e+"\n");
					}
				}
			} else { // 0.2.50 or before
				for (var i=0; i<itemTB.length; i++) {
					var itemName = itemTB[i];
					if (itemName.indexOf("-") == 0) continue; //obsolete item
					try {
						itemName = this.convertTB2PropertyToTB3(itemName);
						if (itemValueArray[i]) {
							//abItem[itemTB[i]] = itemValueArray[i].replace(/_gcslash_/g,'/'); //unescape _gcslash_;
							if (abItem instanceof Components.interfaces.nsIAbCard) {
								abItem.setProperty(itemName, itemValueArray[i].replace(/_gcslash_/g,'/')); //unescape _gcslash_;
							} else {
								abItem[itemTB[i]] = itemValueArray[i].replace(/_gcslash_/g,'/'); //unescape _gcslash_;
							}
						} else {
							//abItem[itemTB[i]] = "";
							if (abItem instanceof Components.interfaces.nsIAbCard) {
								abItem.setProperty(itemName, "");
							} else {
								abItem[itemTB[i]] = "";
							}
						}
					} catch(e) {
						gGCLogger.dump(e+"\n");
					}
				}
			}
		}
	}
}

GCXMLEntry.prototype.convertTB2PropertyToTB3 = function(propName)
{
	return propName;
}

//Group
function GCGroupEntry(entry)
{
	var superConstructor = GCXMLEntry;
	superConstructor.apply(this, arguments);
	this.mailList = null;
	this.listCard = null;
	this.dumpstr = "";
	this.systemDispName = null;
	this.suffix = "";
	
	if (!entry) this.createXMLEntry();
	
	this.itemObj = {
		listItem: ["description", "listNickName"]
	};

	this.initSystemGroups();
}

//Inherit
GCClassInheriter.inherit(GCGroupEntry, GCXMLEntry);

GCGroupEntry.prototype.initSystemGroups = function()
{
	if (this.isSystemGroup()) {
		var name = this.entry.getElementsByTagNameNS(this.atom, "title")[0].firstChild.nodeValue;
		var groupId = this.getSystemGroupId();
		//switch (name) {
		switch (groupId) {
			//case "System Group: My Contacts":
			case "Contacts":
				try {
					this.systemDispName = this.localeBundle.getString("gmcont.system_group_mycontacts");
				} catch(e) {
					this.systemDispName = name;
				}
				break;
			//case "System Group: Friends":
			case "Friends":
				try {
					this.systemDispName = this.localeBundle.getString("gmcont.system_group_friends");
				} catch(e) {
					this.systemDispName = name;
				}
				break;
			//case "System Group: Family":
			case "Family":
				try {
					this.systemDispName = this.localeBundle.getString("gmcont.system_group_family");
				} catch(e) {
					this.systemDispName = name;
				}
				break;
			//case "System Group: Coworkers":
			case "Coworkers":
				try {
					this.systemDispName = this.localeBundle.getString("gmcont.system_group_coworkers");
				} catch(e) {
					this.systemDispName = name;
				}
				break;
			default:
				gGCLogger.dump("Unknow system group: "+name+"\n");
				this.systemDispName = name;
				break;
		}
	}
}

GCGroupEntry.prototype.createXMLEntry = function()
{
	this.shallowDumpStr = "";
	var parser = new DOMParser();
	var entryStr = 
		'<atom:entry xmlns:atom="'+this.atom+'" '
		+ 'xmlns:gd="'+this.gd+'" '
		+ 'xmlns:openSearch="'+this.openSearch+'" '
		+ 'xmlns:gContact="'+this.gContact+'"/>';
	var dom = parser.parseFromString(entryStr, "text/xml");
	if (dom.documentElement.nodeName == "parsererror") {
		this.entry = null;
		gGCLogger.dump("Failed to create Group XML\n");
	} else {
		this.entry = dom.documentElement;
		var catNode = document.createElementNS(this.atom, "category");
		catNode.setAttribute("scheme", "http://schemas.google.com/g/2005#kind");
  	catNode.setAttribute("term", "http://schemas.google.com/contact/2008#group");
  	this.entry.appendChild(catNode);
		var titleElem = document.createElementNS(this.atom, "title");
		var textNode = document.createTextNode(" ");
		titleElem.appendChild(textNode);
		this.entry.appendChild(titleElem);
	}
}

GCGroupEntry.prototype.createMailList = function(addrBookDir)
{
  var mailList = Components.classes["@mozilla.org/addressbook/directoryproperty;1"].createInstance();
  mailList = mailList.QueryInterface(Components.interfaces.nsIAbDirectory);
  mailList.isMailList = true;
  this.updateMailListFromGGroup(mailList);
	//addrBookDir.addMailList(mailList);
  return mailList;
}

GCGroupEntry.prototype.updateMailListFromGGroup = function(list, card)
{
	if (!list) return;
	var listName = this.getTitle();
	//list.dirName = listName+this.suffix;
	list.dirName = listName;
	list.listNickName = "";  
	list.description = ""; 
	//if (card) card.setProperty("DisplayName", listName+this.suffix);
	if (card) card.setProperty("DisplayName", listName);
	
	this.loadTBOnlyItemsFromGContact(list);
	
	this.dumpstr = this.dump(list);
	this.shallowDumpStr = this.dump(list, true);
}

GCGroupEntry.prototype.updateGGroupFromMailList = function(list)
{
//	if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return;
	if (!this.isSystemGroup()) {
		var listName = list.dirName;
		if (this.suffix) {
			var re = new RegExp(this.suffix + "$");
			listName = listName.replace(re, "");
		}
		var titleElem = this.entry.getElementsByTagNameNS(this.atom, "title");
		if (titleElem) {
			titleElem[0].firstChild.nodeValue = listName;
		} else {
			var newElem = document.createElementNS(this.atom, "title");
			var newNode = document.createTextNode(listName);
			newElem.appendChild(newNode);
			this.entry.appendChild(newElem);
		}
	}
	this.setTBOnlyItemsToGContact(list);
	this.dumpstr = this.dump(list);
	this.shallowDumpStr = this.dump(list, true);
}

GCGroupEntry.prototype.setMailList = function(list, account)
{
	if (!list) return;
	this.mailList = list;
	this.dumpstr = this.dump(list);
	this.shallowDumpStr = this.dump(list, true);
	var uri = null;
	try {
		uri = list.URI;
	} catch(e) {
		uri = null;
	}
	if (account && uri && this.getId()) {
		nsPreferences.setUnicharPref("gmcont."+account+"."+uri+".groupId", this.getId());
	}
}

GCGroupEntry.prototype.setListCard = function(card)
{
	this.listCard = card;
}

GCGroupEntry.prototype.getMailList = function()
{
	return this.mailList;
}

GCGroupEntry.prototype.getListCard = function(card)
{
	return this.listCard;
}

GCGroupEntry.prototype.clearMailListChildren = function()
{
	try {
		this.mailList.addressLists.clear();
	} catch(e) {
		var length = this.mailList.addressLists.length;
		for (var i = length - 1; i>=0; i--) {
			try {
		    this.mailList.addressLists.removeElementAt(i);
	    } catch(e2) {
	    }
		}
	}
}

GCGroupEntry.prototype.isChanged = function(list, woList)
{
	var newDumpStr = this.dump(list, woList);
	gGCLogger.dump(this.dumpstr+"\n");
	gGCLogger.dump(newDumpStr+"\n");
	
	return woList ? this.shallowDumpStr != newDumpStr : this.dumpstr != newDumpStr; 
}

GCGroupEntry.prototype.setSuffix = function(suffix)
{
	this.suffix = suffix;
}

GCGroupEntry.prototype.getTitle = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var title = "";
	if (this.isSystemGroup()) {
		title = this.systemDispName;
	} else {
		title = this.entry.getElementsByTagNameNS(this.atom, "title")[0].firstChild.nodeValue + this.suffix;
	}
	return title;
}

GCGroupEntry.prototype.getDispName = function()
{
	return this.getTitle();
}

GCGroupEntry.prototype.getSystemDispName = function()
{
	return this.systemDispName;
}

GCGroupEntry.prototype.getSystemName = function()
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return "";
	return this.entry.getElementsByTagNameNS(this.atom, "title")[0].firstChild.nodeValue;
}

GCGroupEntry.prototype.getOriginalName = function()
{
	return this.getSystemName();
}

GCGroupEntry.prototype.getSystemGroupId = function()
{
//	if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return null;
	var ret = null
	try {
		ret =  this.entry.getElementsByTagNameNS(this.gContact, "systemGroup")[0].getAttribute("id");
	} catch(e) {
		ret = null;
	}
	
	return ret;
}

GCGroupEntry.prototype.isSystemGroup = function()
{
	if (!this.entry && !this.loadCache()) return false;

	if (this.getSystemGroupId()) return true;
	else return false;
}

GCGroupEntry.prototype.dump = function(list, woList)
{
	var name = list.dirName ? list.dirName : "";
	if (name && this.suffix) {
		var re = new RegExp(this.suffix + "$");
		name = name.replace(re,"");
	}
	var desc = list.description ? list.description : "";
	var nickName = list.listNickName ? list.listNickName : "";
	var addrsStr = "";
	var addrsArray = new Array();
	if (!woList && list.addressLists) {
		var total = list.addressLists.length;
    if (total) {
      for ( var i = 0;  i < total; i++ ) {
      	var card = list.addressLists.queryElementAt(i, Components.interfaces.nsIAbCard)
        addrsArray.push(card.getProperty("PrimaryEmail", ""));
      }
    }
    addrsStr = addrsArray.sort().join("");
  }
  return name + desc + nickName + addrsStr;
}

//Contact
function GCContactEntry(entry)
{
	var superConstructor = GCXMLEntry;
	superConstructor.apply(this, arguments);
	this.card = null;
	this.dumpstr = "";
	this.photoManager = null;
	if (!entry) this.createXMLEntry();
	
	//update items only in TB
	this.itemObj = {
		contactTB: [
			//Contact
			"-FirstName",
			"-LastName",
			"-NickName",
			"-SecondEmail",
			"-aimScreenName", //aimScreenName is not a TB only item. Rename for compatibility
			"PreferMailFormat",
			"AllowRemoteContent",
			"-PhoneticFirstName",
			"-PhoneticLastName",
			"PreferDisplayName"
		],
		homeAddrTB: [
			//Address - Home
			"HomeAddress2",
			"HomeCity",
			"HomeState",
			"HomeCountry",
			"HomeZipCode",
			"-WebPage2",
			"-Birthday",
			"-BirthMonth",
			"-BirthYear",
			"-Age"
		],
		workAddrTB: [
			//Address - Work
			"-Department",
			"-Company",
			"WorkAddress2",
			"WorkCity",
			"WorkState",
			"WorkZipCode",
			"WorkCountry",
			"-WebPage1"
		],
		otherTB: [
			//Other
			"-Custom1",
			"-Custom2",
			"-Custom3",
			"-Custom4"
		]
	};
}

//Inherit
GCClassInheriter.inherit(GCContactEntry, GCXMLEntry);

GCContactEntry.pseudoEmailsCount = 0;

GCContactEntry.prototype.createXMLEntry = function()
{
	var parser = new DOMParser();
	var entryStr = 
		'<atom:entry xmlns:atom="' + this.atom
		+ '" xmlns:gd="' + this.gd
		+ '" xmlns:openSearch="' + this.openSearch
		+ '" xmlns:gContact="' + this.gContact + '"/>';
	var dom = parser.parseFromString(entryStr, "text/xml");
	
	if (dom.documentElement.nodeName == "parsererror") {
		this.entry = null;
		gGCLogger.dump("Failed to create Entry XML\n");
	} else {
		this.entry = dom.documentElement;
		var catNode = document.createElementNS(this.atom, "category");
		catNode.setAttribute("scheme", "http://schemas.google.com/g/2005#kind");
  	catNode.setAttribute("term", "http://schemas.google.com/contact/2008#contact");
  	this.entry.appendChild(catNode);
	}
}

GCContactEntry.prototype.getABCard = function()
{
	return this.card;
}

GCContactEntry.prototype.setABCard = function(card)
{
	this.card = card;
	//this.dumpstr = this.card.convertToEscapedVCard();
	this.dumpstr = this.dump(this.card);
}

GCContactEntry.prototype.createABCard = function(addrBookDir)
{
	var card = Components.classes["@mozilla.org/addressbook/cardproperty;1"].
								createInstance(Components.interfaces.nsIAbCard);
	this.updateCardFromGContact(card);
	var abCard = addrBookDir.addCard(card);
	return abCard;
}

GCContactEntry.prototype.belongToGroup = function(gcGroup)
{
	//if (!this.entry) this.loadCache();
	if (!this.entry && !this.loadCache()) return;
	var id = gcGroup instanceof GCGroupEntry ? gcGroup.getId() : gcGroup;

	//check if this contact belongs to the group
	var groups = this.entry.getElementsByTagNameNS(this.gContact, "groupMembershipInfo");
	for (var i=0; i<groups.length; i++) {
		if (groups[i].getAttribute("href") == id) return; //already added
	}

	var xmlGrp = document.createElementNS(this.gContact, "groupMembershipInfo");
	xmlGrp.setAttribute("href", id);
	this.entry.appendChild(xmlGrp);
	gGCLogger.dump("Belong to group\n");
}

GCContactEntry.prototype.unbelongToGroup = function(gcGroup)
{
	if (!this.entry && !this.loadCache()) return;
	var id = gcGroup.getId();
	var group = null;
	var groups = this.entry.getElementsByTagNameNS(this.gContact, "groupMembershipInfo");
	for (var i=0; i<groups.length; i++) {
		var group = groups[i];
		if (group.getAttribute("href") == id) {
			this.entry.removeChild(group);
			return;
		}
	}
}

GCContactEntry.prototype.getBelongingGroupIds = function()
{
	var ids = [];
	if (!this.entry && !this.loadCache()) return ids;
	var groups = this.entry.getElementsByTagNameNS(this.gContact, "groupMembershipInfo");
	for (var i=0; i<groups.length; i++) {
		var group = groups[i];
		var id = group.getAttribute("href");
		if (id) ids.push(id);
	}
	
	return ids;
}

GCContactEntry.prototype.setPhotoManager = function(manager)
{
	this.photoManager = manager;
}

GCContactEntry.prototype.clearPhotoManager = function()
{
	this.photoManager = null;
}

GCContactEntry.prototype.setPhoto = function(photoURI)
{
  this.card.setProperty("PhotoType", "file");
  // Store the original URI
  this.card.setProperty("PhotoURI", photoURI);
  // Save the new file and store its URI as PhotoName
  var file = null;
  try {
  	//TB7 or later
  	file = this.photoManager.storePhoto(photoURI);
  } catch(e) {
  	file = this.photoManager.savePhoto(photoURI);
  }
  if (file) {
		// Remove the original, if any
    this.photoManager.removePhoto(this.card.getProperty("PhotoName", null));
    this.card.setProperty("PhotoName", file.leafName);
  } else {
  	gGCLogger.dump("save photo failed\n");
  }
	return this.card;
}

GCContactEntry.prototype.resetPhoto = function()
{
  try {
  	this.card.setProperty("PhotoType", "generic");
  	this.card.setProperty("PhotoURI", "");
  	this.photoManager.removePhoto(this.card.getProperty("PhotoName", null));
  	this.card.setProperty("PhotoName", "");
  }catch(e){
  	gGCLogger.dump(e);
  }
  
  return this.card;
}

GCContactEntry.prototype.updateCardFromGContact = function(abCard)
{
	if (!this.entry && !this.loadCache()) return false;
	abCard.QueryInterface(Components.interfaces.nsIAbCard);

	var oldEtag = abCard.getProperty("GmContactEtag",null);
	var newEtag = this.getEntryEtag();
	if (!nsPreferences.getBoolPref("gmcont.use_contact_etag", false)) oldEtag = null;
	if (oldEtag && newEtag && oldEtag == newEtag) {
		this.dumpstr = this.dump(abCard);
		return false;
	}
//move to the bottom to refresh etag only when all updates are succeeded
//	else if (newEtag) {
//		abCard.setProperty("GmContactEtag", newEtag);
//	}

	//Contact tab
	//Full Name/Display Name
	var fullNameElem = this.entry.getElementsByTagNameNS(this.gd, "fullName")[0];
	if (fullNameElem && fullNameElem.firstChild) abCard.setProperty("DisplayName", fullNameElem.firstChild.nodeValue);
	else abCard.setProperty("DisplayName", "");

	//Given Name/First Name
	var givenNameElem = this.entry.getElementsByTagNameNS(this.gd, "givenName")[0];
	if (givenNameElem && givenNameElem.firstChild) abCard.setProperty("FirstName", givenNameElem.firstChild.nodeValue);
	else abCard.setProperty("FirstName", "");

	//Family Name/Last Name
	var familyNameElem = this.entry.getElementsByTagNameNS(this.gd, "familyName")[0];
	if (familyNameElem && familyNameElem.firstChild) abCard.setProperty("LastName", familyNameElem.firstChild.nodeValue);
	else abCard.setProperty("LastName", "");

	//Yomi
	var givenNameYomi = givenNameElem ? givenNameElem.getAttribute("yomi") : "";
	abCard.setProperty("PhoneticFirstName", givenNameYomi);
	var familyNameYomi = familyNameElem ? familyNameElem.getAttribute("yomi") : "";
	abCard.setProperty("PhoneticLastName", familyNameYomi);

	//Nickname
	var nickNameElem = this.entry.getElementsByTagNameNS(this.gContact, "nickname")[0];
	if (nickNameElem && nickNameElem.firstChild) abCard.setProperty("NickName", nickNameElem.firstChild.nodeValue);
	else abCard.setProperty("NickName", "");

	//email
	//primary email
	var emails = this.entry.getElementsByTagNameNS(this.gd, "email");
	var primEmail = "";
	var secondEmail = "";
	for (var i=0; i<emails.length; i++) {
		var email = emails[i];
		if (!primEmail && email.getAttribute("primary") == "true") primEmail = email.getAttribute("address");
		else if (!secondEmail && !email.getAttribute("primary")) secondEmail = email.getAttribute("address");
		
		if (primEmail && secondEmail) break;
	}
	
	//primary email was not found
	if (!primEmail && emails.length > 0) {
		primEmail = emails[0].getAttribute("address");
		secondEmail = emails[1] && emails[1].getAttribute("address") ? emails[1].getAttribute("address") : "";
	}
	abCard.setProperty("PrimaryEmail", primEmail);
	abCard.setProperty("SecondEmail", secondEmail);
	
	var addPseudoEmail = nsPreferences.getBoolPref("gmcont.add_pseudo_email", false);
	if (addPseudoEmail && !primEmail) {
		var id = this.getId().split("/").pop();
		GCContactEntry.pseudoEmailsCount++;
		abCard.setProperty("PrimaryEmail", id+"@pseudo.googlecontacts");
	}
	
	//Chat
	var im = [];
	im["GOOGLE_TALK"] = "_GoogleTalk";
	im["AIM"] = "_AimScreenName";
	im["YAHOO"] = "_Yahoo";
	im["SKYPE"] = "_Skype";
	im["QQ"] = "_QQ";
	im["MSN"] = "_MSN";
	im["ICQ"] = "_ICQ";
	im["JABBER"] =  "_JabberId";
	var imElems = this.entry.getElementsByTagNameNS(this.gd, "im");
	for (var i=0; i<imElems.length; i++) {
		var protocol = imElems[i].getAttribute("protocol") ? imElems[i].getAttribute("protocol").split("#")[1] : null;
		var account = imElems[i].getAttribute("address");
		if (protocol && im[protocol]) {
			abCard.setProperty(im[protocol], account);
			im[protocol] = null;
		}
	}
	
	var key = null;
	for (key in im) {
		if (im[key]) abCard.setProperty(im[key], "");
	}
	
	//phone number
	var mainPhoneMap = nsPreferences.copyUnicharPref("gmcont.main_phone_map","none");
	var workPhoneAttribute = mainPhoneMap == "WorkPhone" ? "main" : "work";
	var homePhoneAttribute = mainPhoneMap == "HomePhone" ? "main" : "home";
	var mobilePhoneAttribute = mainPhoneMap == "MobilePhone" ? "main" : "mobile";
	var faxAttribute = nsPreferences.copyUnicharPref("gmcont.fax_attribute","home_fax");
	var workPhone = "";
	var homePhone = "";
	var fax = "";
	var pager = "";
	var cell = "";
	
	var phones = this.entry.getElementsByTagNameNS(this.gd, "phoneNumber");
	for (var i=0; i<phones.length; i++) {
		var rel = phones[i].getAttribute("rel");
		var textNode = phones[i].firstChild;
		if (rel && textNode) {
			if (!workPhone && rel.split("#")[1] == workPhoneAttribute) workPhone = textNode.nodeValue;
			else if (!homePhone && rel.split("#")[1] == homePhoneAttribute) homePhone = textNode.nodeValue;
			else if (!fax && rel.split("#")[1] == faxAttribute) fax = textNode.nodeValue;
			else if (!pager && rel.split("#")[1] == "pager") pager = textNode.nodeValue;
			else if (!cell && rel.split("#")[1] == mobilePhoneAttribute) cell = textNode.nodeValue;
		}
	}
	abCard.setProperty("WorkPhone", workPhone);
	abCard.setProperty("HomePhone", homePhone);
	abCard.setProperty("FaxNumber", fax);
	abCard.setProperty("PagerNumber", pager);
	abCard.setProperty("CellularNumber", cell);

	this.loadTBOnlyItemsFromGContact(abCard);

	//Address tab
	//home
	var addrFields = [];
	addrFields = this.parseGContactAddrXML("Home");
	var key;
	for (key in addrFields) {
		var field = addrFields[key];
		if (field) {
			abCard.setProperty(key, field);
		} else {
			abCard.setProperty(key, "");
		}
	}
	
	//work
	addrFields = [];
	addrFields = this.parseGContactAddrXML("Work");
	for (key in addrFields) {
		var field = addrFields[key];
		if (field) {
			abCard.setProperty(key, field);
		} else {
			abCard.setProperty(key, "");
		}
	}
	
	//Organization
	var org = this.entry.getElementsByTagNameNS(this.gd, "organization")[0];
	var jobTitle = "";
	var company = "";
	var department = "";
	if (org) {
		var titleElem = org.getElementsByTagNameNS(this.gd,"orgTitle")[0];
		if (titleElem && titleElem.firstChild) jobTitle = titleElem.firstChild.nodeValue;
		var companyElem = org.getElementsByTagNameNS(this.gd, "orgName")[0];
		if (companyElem && companyElem.firstChild) company = companyElem.firstChild.nodeValue;
		var departElem = org.getElementsByTagNameNS(this.gd, "orgDepartment")[0];
		if (departElem && departElem.firstChild) department = departElem.firstChild.nodeValue;
	}
	abCard.setProperty("JobTitle", jobTitle);
	abCard.setProperty("Company", company);
	abCard.setProperty("Department", department);

	//birthday
	var birthElem = this.entry.getElementsByTagNameNS(this.gContact, "birthday")[0];
	var dd = "";
	var mm = "";
	var yyyy = "";
	if (birthElem) {
		var birthdayFields = birthElem.getAttribute("when").split("-");
		dd = parseInt(birthdayFields.pop(), 10);
		mm = parseInt(birthdayFields.pop(), 10);
		var year = birthdayFields.pop();
		if (year) yyyy = parseInt(year, 10);
	}
	abCard.setProperty("BirthYear", yyyy);
	abCard.setProperty("BirthMonth", mm);
	abCard.setProperty("BirthDay", dd);

	//web
	var webs = this.entry.getElementsByTagNameNS(this.gContact, "website");
	var workWeb = "";
	var homeWeb = "";
	for (var i=0; i<webs.length; i++) {
		var rel = webs[i].getAttribute("rel");
		if (!workWeb && rel == "work") workWeb = webs[i].getAttribute("href");
		else if (!homeWeb && rel == "home") homeWeb = webs[i].getAttribute("href");
	}
	abCard.setProperty("WebPage1", workWeb);
	abCard.setProperty("WebPage2", homeWeb);
	
	//Other tab
	//custom1
	var custom1 = "";
	var custom2 = "";
	var custom3 = "";
	var custom4 = "";
	
	var customs = this.entry.getElementsByTagNameNS(this.gContact, "userDefinedField");
	for (var i=0; i<customs.length; i++) {
		var key = customs[i].getAttribute("key"); 
		if (!custom1 && key == "Custom1") custom1 = customs[i].getAttribute("value"); 
		else if (!custom2 && key == "Custom2") custom2 = customs[i].getAttribute("value"); 
		else if (!custom3 && key == "Custom3") custom3 = customs[i].getAttribute("value");
		else if (!custom4 && key == "Custom4") custom4 = customs[i].getAttribute("value"); 
	}
	abCard.setProperty("Custom1", custom1);
	abCard.setProperty("Custom2", custom2);
	abCard.setProperty("Custom3", custom3);
	abCard.setProperty("Custom4", custom4);

	//notes
	var noteElem = this.entry.getElementsByTagNameNS(this.atom, "content")[0];
	if (noteElem && noteElem.firstChild) abCard.setProperty("Notes", noteElem.firstChild.nodeValue);
	else abCard.setProperty("Notes", "");

	this.dumpstr = this.dump(abCard);

	//update etag
	if (newEtag) {
		abCard.setProperty("GmContactEtag", newEtag);
	}
	
	return true;
}

GCContactEntry.prototype.parseGContactAddrXML = function(type) 
{
	var sAddr = null;

	if (!type) return [];
	var addrElems = this.entry.getElementsByTagNameNS(this.gd, "structuredPostalAddress");
	var rel = "http://schemas.google.com/g/2005#" + type.toLowerCase();
	for (var i=0; i<addrElems.length; i++) {
		if (rel == addrElems[i].getAttribute("rel")) {
			sAddr = addrElems[i];
			break;
		}
	}

	var fields = [];
	fields[type+"Address"] = "";
	fields[type+"Address2"] = "";
	fields[type+"City"] = "";
	fields[type+"State"] = "";
	fields[type+"ZipCode"] = "";
	fields[type+"Country"] = "";
	var formattedAddress = "";
	if (sAddr) {
		var street = sAddr.getElementsByTagNameNS(this.gd, "street")[0];
		if (street && street.firstChild) fields[type+"Address"] = street.firstChild.nodeValue.replace(/[\n\r]+/gm, ", ");

		var addr2Pref = nsPreferences.copyUnicharPref("gmcont.address2_fields_for", "none");
		var addr2 = null;
		if (addr2Pref == "po") addr2 = sAddr.getElementsByTagNameNS(this.gd, "pobox")[0];
		else if (addr2Pref == "neighbor") addr2 = sAddr.getElementsByTagNameNS(this.gd, "neighborhood")[0];
		if (addr2 && addr2.firstChild) fields[type+"Address2"] = addr2.firstChild.nodeValue;
		else fields[type+"Address2"] = "";

		var keys = [["City", "city"],
								["State", "region"],
								["ZipCode", "postcode"],
								["Country", "country"]];

		for (var i=0; i<keys.length; i++) {
			var elem = sAddr.getElementsByTagNameNS(this.gd, keys[i][1])[0];
			if (elem && elem.firstChild) fields[type+keys[i][0]] = elem.firstChild.nodeValue;
		}
		var formattedAddressElem = sAddr.getElementsByTagNameNS("gd", "formattedAddress")[0];
		if (formattedAddressElem && formattedAddressElem.firstChild) formattedAddress = formattedAddressElem.firstChild.nodeValue.replace(/[\n\r]+/gm, ", ");
	}
	
	//check if some of fields have values
	var key;
	var filled = false;
	for (key in fields) {
		if (fields[key]) {
			filled = true;
			break;
		}
	}
	
	//only formatted address
	if (!filled) {
		fields[type+"Address"] = formattedAddress;
	}
	
	return fields;
}

GCContactEntry.prototype.parseGContactAddr = function(addr, type) 
{
	var fields = new Array();
	var format = nsPreferences.copyUnicharPref("gmcont.addr_conv_format", "");
	var delimiter = nsPreferences.copyUnicharPref("gmcont.addr_conv_delimiter", ", ");
	if (delimiter == '<BR>') delimiter = "\n";
	
	if (format && type) {
		var items = format.split(",");
		var addrFields = addr ? addr.split(delimiter) : [];
		for (var i=0; i<items.length; i++) {
			var val = addrFields[i] ? addrFields[i] : "";
			var field = {
				prop: type+items[i],
				value: val
			};
			fields.push(field);
		}
	} else if (!format && type) { //all in Address
			var field = {
				prop: type+"Address",
				value: addr
			};
			fields.push(field);
	}
	
	return fields;
}

GCContactEntry.prototype.populateGContactName = function(abCard)
{
	var fullName = abCard.getProperty("DisplayName", null);
	var givenName = abCard.getProperty("FirstName", null);
	var familyName = abCard.getProperty("LastName", null);
	var givenNameYomi = abCard.getProperty("PhoneticFirstName", null);
	var familyNameYomi = abCard.getProperty("PhoneticLastName", null);
	var nameElem = this.entry.getElementsByTagNameNS(this.gd, "name")[0];

	//create name element 
	if (!nameElem && (fullName || givenName || familyName || givenNameYomi || familyNameYomi)) {
		var newElem = document.createElementNS(this.gd, "name");
		this.entry.appendChild(newElem);
		nameElem = this.entry.getElementsByTagNameNS(this.gd, "name")[0];
	}
	
	if (!nameElem) return;
	
	var items = [
				["fullName", fullName, null],
				["givenName", givenName, givenNameYomi],
				["familyName", familyName, familyNameYomi]
	];
	
	for (var i=0; i<items.length; i++) {
		var tag = items[i][0];
		var val = items[i][1];
		var yomi = items[i][2];
		var elem = nameElem.getElementsByTagNameNS(this.gd, tag)[0];
		if (val) {
			if (elem && elem.firstChild) {
				elem.firstChild.nodeValue = val;
				if (yomi) elem.setAttribute("yomi", yomi);
				else elem.removeAttribute("yomi");
			} else {
				elem = document.createElementNS(this.gd, tag);
				var newNode = document.createTextNode(val);
				if (yomi) elem.setAttribute("yomi", yomi);
				elem.appendChild(newNode);
				nameElem.appendChild(elem);
			}
		} else if (elem) {
			nameElem.removeChild(elem);
		}
	}
	
	var nickName = abCard.getProperty("NickName", null);
	var nickElem = this.entry.getElementsByTagNameNS(this.gContact, "nickname")[0];
	if (nickName) {
		if (nickElem && nickElem.firstChild) {
			nickElem.firstChild.nodeValue = nickName;
		} else {
			nickElem = document.createElementNS(this.gContact, "nickname");
			var newNode = document.createTextNode(nickName);
			nickElem.appendChild(newNode);
			this.entry.appendChild(nickElem);
		}
	} else if (nickElem) {
		this.entry.removeChild(nickElem);
	}
}

GCContactEntry.prototype.populateGContactEmail = function(abCard)
{
	var pAddrElem = null;
	var sAddrElem = null;
	var emailElems = this.entry.getElementsByTagNameNS(this.gd, "email");
	if (emailElems) {
		pAddrElem = emailElems[0];
		sAddrElem = emailElems[1];
	}
	
	//find primary email
	if (pAddrElem && !pAddrElem.getAttribute("primary")) {
		for (var i=1; i<emailElems.length; i++) {
			if (emailElems[i].getAttribute("primary")) {
				pAddrElem = emailElems[i];
				sAddrElem = emailElems[0];
				break;
			}
		}
	}
	
	var delPrimary = false;
	var pAddr = abCard.getProperty("PrimaryEmail", null);
	if (pAddr && pAddr.indexOf("@pseudo.googlecontacts") != -1) pAddr = "pseudo";
	if (pAddr && pAddr != "pseudo") {
		if (pAddrElem) {
			pAddrElem.setAttribute("address", pAddr);
		} else {
			pAddrElem = document.createElementNS(this.gd, "email");
			pAddrElem.setAttribute("address", pAddr)
			pAddrElem.setAttribute("primary", "true")
			pAddrElem.setAttribute("rel", "http://schemas.google.com/g/2005#other");
			this.entry.appendChild(pAddrElem);
		}
	} else if (pAddrElem) {
		this.entry.removeChild(pAddrElem)
		delPrimary = true;
	}
	if (!pAddr){
		var addPseudoEmail = nsPreferences.getBoolPref("gmcont.add_pseudo_email", false);
		if (addPseudoEmail) {
			GCContactEntry.pseudoEmailsCount++;
			abCard.setProperty("PrimaryEmail",
													GCContactEntry.pseudoEmailsCount+"@pseudo.googlecontacts");
		}
		delPrimary = true;
	}

	var sAddr = abCard.getProperty("SecondEmail", null);
	if (sAddr) {
		if (sAddrElem) {
			sAddrElem.setAttribute("address", sAddr);
		} else {
			sAddrElem = document.createElementNS(this.gd, "email");
			sAddrElem.setAttribute("address", sAddr)
			sAddrElem.setAttribute("rel", "http://schemas.google.com/g/2005#other");
			this.entry.appendChild(sAddrElem);
		}
	} else if (sAddrElem){
		this.entry.removeChild(sAddrElem)
	}

	emailElems = this.entry.getElementsByTagNameNS(this.gd, "email");
	if (delPrimary) {
		//secondary email should be primary
		if (sAddr) {
			gGCLogger.dump("No primary email. Secondary one becomes a primary.\n")
			emailElems[0].setAttribute("primary", "true");
			abCard.setProperty("PrimaryEmail", sAddr);
			//set new secondary
			gGCLogger.dump("Set new secondary email to a card.\n")
			if (emailElems[1]) {
				abCard.setProperty("SecondEmail", emailElems[1].getAttribute("address"));
			} else {
				abCard.setProperty("SecondEmail", "");
			}
		}
	}
}

GCContactEntry.prototype.populateGContactIM = function(abCard)
{
	//im = [GC Name] = [TB Name, GC elem]
	var im = [];
	im["GOOGLE_TALK"] = ["_GoogleTalk", null];
	im["AIM"] = ["_AimScreenName", null];
	im["YAHOO"] = ["_Yahoo", null];
	im["SKYPE"] = ["_Skype", null];
	im["QQ"] = ["_QQ", null];
	im["MSN"] = ["_MSN", null];
	im["ICQ"] = ["_ICQ", null];
	im["JABBER"] =  ["_JabberId", null];

	var imElems = this.entry.getElementsByTagNameNS(this.gd, "im");
	for (var i=0; i<imElems.length; i++) {
		var protocol = imElems[i].getAttribute("protocol") ? imElems[i].getAttribute("protocol").split("#")[1] : null;
		if (protocol && !im[protocol][1]) im[protocol][1] = imElems[i];
		//var account = imElems[i].getAttribute("address");
	}

	var key = null;
	for (key in im) {
		var val = abCard.getProperty(im[key][0], null);
		if (val) {
			if (im[key][1]) {
				im[key][1].setAttribute("address", val);
			} else {
				var elem = document.createElementNS(this.gd, "im");
				elem.setAttribute("address", val);
				elem.setAttribute("protocol", "http://schemas.google.com/g/2005#"+key);
				elem.setAttribute("rel", "http://schemas.google.com/g/2005#other");
				this.entry.appendChild(elem);
			}
		} else if (im[key][1]) {
			this.entry.removeChild(im[key][1])
		}
	}
}

GCContactEntry.prototype.populateGContactPhone = function(abCard)
{
	var mainPhoneMap = nsPreferences.copyUnicharPref("gmcont.main_phone_map","none");
	var workPhoneAttribute = mainPhoneMap == "WorkPhone" ? "main" : "work";
	var homePhoneAttribute = mainPhoneMap == "HomePhone" ? "main" : "home";
	var mobilePhoneAttribute = mainPhoneMap == "MobilePhone" ? "main" : "mobile";
	var faxAttribute = nsPreferences.copyUnicharPref("gmcont.fax_attribute","home_fax");
	if (faxAttribute != "home_fax" && faxAttribute != "work_fax") faxAttribute = "home_fax";
	
	var phoneType = [];
	phoneType[workPhoneAttribute] = ["WorkPhone", null];
	phoneType[homePhoneAttribute] = ["HomePhone", null];
	phoneType[faxAttribute] = ["FaxNumber", null];
	phoneType["pager"] = ["PagerNumber", null];
	phoneType[mobilePhoneAttribute] = ["CellularNumber", null];
	
	var phoneElems = this.entry.getElementsByTagNameNS(this.gd, "phoneNumber");
	for (var i=0; i<phoneElems.length; i++) {
		var attr = phoneElems[i].getAttribute("rel") ? phoneElems[i].getAttribute("rel").split("#")[1] : null;
		if (attr && phoneType[attr] && !phoneType[attr][1]) phoneType[attr][1] = phoneElems[i];
	}
	
	var key = null;
	for (key in phoneType) {
		var val = abCard.getProperty(phoneType[key][0], null);
		if (val) {
			if (phoneType[key][1] && phoneType[key][1].firstChild) {
				phoneType[key][1].firstChild.nodeValue = val;
			} else {
				var elem = document.createElementNS(this.gd, "phoneNumber");
				elem.setAttribute("rel", "http://schemas.google.com/g/2005#" + key);
				var newNode = document.createTextNode(val);
				elem.appendChild(newNode);
				this.entry.appendChild(elem);
			}
		} else if (phoneType[key][1]) {
			this.entry.removeChild(phoneType[key][1])
		}
	}
}

GCContactEntry.prototype.populateGContactOrg = function(abCard)
{
	var orgElems = this.entry.getElementsByTagNameNS(this.gd, "organization");
	var orgElem = null;
	for (var i=0; i<orgElems.length; i++) {
		var rel = orgElems[i].getAttribute("rel")
		if (rel && rel.split("#")[1] == "other") {
			orgElem = orgElems[i];
			break;
		}
	}
	
	var jTitle = abCard.getProperty("JobTitle", null);
	var cmp = abCard.getProperty("Company", null);
	var department = abCard.getProperty("Department", null);
	
	if (!jTitle && !cmp && !department) {
		if (orgElem) this.entry.removeChild(orgElem);
		return;
	}

	var tags = ["orgTitle", "orgName", "orgDepartment"];
	var vals = [jTitle, cmp, department];
	for (var i=0; i<tags.length; i++) {
		var tag = tags[i];
		var val = vals[i];
		var elem = orgElem.getElementsByTagNameNS(this.gd, tag)[0];
		if (val) {
			if (elem && elem.firstChild) {
				elem.firstChild.nodeValue = val;
			} else {
				elem = document.createElementNS(this.gd, tag);
				var textNode = document.createTextNode(val);
				elem.appendChild(textNode);
				orgElem.appendChild(elem);
			}
		} else if (elem) {
			orgElem.removeChild(elem);
		}
	}
}

GCContactEntry.prototype.populateGContactBirthday = function(abCard)
{
	var year = abCard.getProperty("BirthYear", null);
	if (!year) year = "-";
	else if (parseInt(year) < 100) year = 1900 + parseInt(year);
	var month = abCard.getProperty("BirthMonth", null);
	var day = abCard.getProperty("BirthDay", null);
	var birthElem = this.entry.getElementsByTagNameNS(this.gContact, "birthday")[0];
	if (month && day) {
		if (!birthElem) {
			birthElem = document.createElementNS(this.gContact, "birthday");
			birthElem.setAttribute("when", year+"-"+month+"-"+day);
			this.entry.appendChild(birthElem);
		}
		birthElem.setAttribute("when", year+"-"+month+"-"+day);
	} else if (birthElem){
		this.entry.removeChild(birthElem);
	}
}

GCContactEntry.prototype.populateGContactWeb = function(abCard)
{
	var webVal1 = abCard.getProperty("WebPage1", null); 
	var webVal2 = abCard.getProperty("WebPage2", null);
	
	var webElems = this.entry.getElementsByTagNameNS(this.gContact, "website");
	var webElem1 = null;
	var webElem2 = null;
	for (var i=0; i<webElems.length; i++) {
		var web = webElems[i];
		if (!webElem1 && web.getAttribute("rel") == "work") webElem1 = web;
		else if (!webElem2 && web.getAttribute("rel") == "home") webElem2 = web;
		if (webElem1 && webElem2) break;
	}
	
	if (webVal1) {
		if (webElem1) {
			webElem1.setAttribute("href", webVal1);
		} else {
			webElem1 = document.createElementNS(this.gContact, "website");
			webElem1.setAttribute("rel", "work");
			webElem1.setAttribute("href", webVal1);
			this.entry.appendChild(webElem1);
		}
	} else if (webElem1) {
		this.entry.removeChild(webElem1);
	}

	if (webVal2) {
		if (webElem2) {
			webElem2.setAttribute("href", webVal2);
		} else {
			webElem2 = document.createElementNS(this.gContact, "website");
			webElem2.setAttribute("rel", "home");
			webElem2.setAttribute("href", webVal2);
			this.entry.appendChild(webElem2);
		}
	} else if (webElem2) {
		this.entry.removeChild(webElem2);
	}
}

GCContactEntry.prototype.populateGContactCustom = function(abCard)
{
	var custVal1 = abCard.getProperty("Custom1", null);
	var custVal2 = abCard.getProperty("Custom2", null);
	var custVal3 = abCard.getProperty("Custom3", null);
	var custVal4 = abCard.getProperty("Custom4", null);
	
	var custElems = this.entry.getElementsByTagNameNS(this.gContact, "userDefinedField");
	var custItems = [];
	custItems["Custom1"] = [custVal1, null];
	custItems["Custom2"] = [custVal2, null];
	custItems["Custom3"] = [custVal3, null];
	custItems["Custom4"] = [custVal4, null];
	
	for (var i=0; i<custElems.length; i++) {
		var custElem = custElems[i];
		var key = custElem.getAttribute("key");
		if (!custItems[key][1]) custItems[key][1] = custElem;
		
		if (custItems["Custom1"][1] && custItems["Custom2"][1] && custItems["Custom3"][1] && custItems["Custom4"][1]) break;
	}
	
	var key = null;
	for (key in custItems) {
		var custItem = custItems[key];
		if (custItem[0]) {
			if (custItem[1]) {
				custItem[1].setAttribute("value", custItem[0]);
			} else {
				var elem = document.createElementNS(this.gContact, "userDefinedField");
				elem.setAttribute("key", key);
				elem.setAttribute("value", custItem[0]);
				this.entry.appendChild(elem);
			}
		} else if (custItem[1]) {
			this.entry.removeChild(custItem[1]);
		}
	}
}

GCContactEntry.prototype.updateGContactFromCard = function(abCard)
{
	if (!this.entry && !this.loadCache()) return false;
	abCard.QueryInterface(Components.interfaces.nsIAbCard);
	if (!abCard.getProperty("DisplayName", null) && !abCard.getProperty("PrimaryEmail", null)) {
		var msg = this.localeBundle.getString("gmcont.no_primary_email");
		alert(msg);
		return false;
	}

	//Contact tab
	//Name
	this.populateGContactName(abCard);

	//email
	this.populateGContactEmail(abCard);

	//Chat
	this.populateGContactIM(abCard);

	//Phone
	this.populateGContactPhone(abCard);

	//Address tab
	this.populateGContactAddr(abCard, "Home");
	this.populateGContactAddr(abCard, "Work");
	
	this.populateGContactOrg(abCard);
	
	//Birthday
	this.populateGContactBirthday(abCard);

	//web
	this.populateGContactWeb(abCard);

	//Other tab
	this.populateGContactCustom(abCard);

	//custom
	var nts = abCard.getProperty("Notes", null);
	var ntsElem = this.entry.getElementsByTagNameNS(this.atom, "content")[0];
	if (nts) {
		if (ntsElem && ntsElem.firstChild) {
			ntsElem.firstChild.nodeValue = nts;
		} else {
			ntsElem = document.createElementNS(this.atom, "content");
			var textNode = document.createTextNode(nts);
			ntsElem.appendChild(textNode);
			this.entry.appendChild(ntsElem)
		}
	} else if (ntsElem) {
		this.entry.removeChild(ntsElem);
	}

	//TB only items
	this.setTBOnlyItemsToGContact(abCard);

	//this.dumpstr = abCard.convertToEscapedVCard();
	this.dumpstr = this.dump(abCard);

	return true;
}

GCContactEntry.prototype.populateGContactAddr = function(abCard, type)
{
	var sAddrNode = null;
	var newNode = false;
	var allNodes = this.entry.getElementsByTagNameNS(this.gd, "structuredPostalAddress");
	var rel = "http://schemas.google.com/g/2005#" + type.toLowerCase();
	for (var i=0; i<allNodes.length; i++) {
		var node = allNodes[i];
		if (node.getAttribute("rel") == rel) {
			sAddrNode = node;
			break;
		}
	}
	
	if (!sAddrNode) {
		sAddrNode = document.createElementNS(this.gd, "structuredPostalAddress");
		sAddrNode.setAttribute("rel", rel);
		newNode = true;
	}

	var fAddrNode = sAddrNode.getElementsByTagNameNS(this.gd, "formattedAddress")[0];
	if (fAddrNode) sAddrNode.removeChild(fAddrNode);

	//keys[TB name] = GC name
	var keys = [];
	keys["Address"] = "street";
	keys["City"] = "city";
	keys["State"] = "region";
	keys["ZipCode"] = "postcode";
	keys["Country"] = "country";
	
	var addr2Pref = nsPreferences.copyUnicharPref("gmcont.address2_fields_for", "none");
	if (addr2Pref == "po") {
		keys["Address2"] = "pobox";
	} else if (addr2Pref == "neighbor") {
		keys["Address2"] = "neighborhood";
	}
	
	var empty = true; //all field are empty = structuredPostalAddress is not needed
	var key = null;
	for (key in keys) {
		var val = abCard.getProperty(type+key, null);
		var domNode = sAddrNode.getElementsByTagNameNS(this.gd, keys[key])[0];
		if (val) {
			if (domNode && domNode.firstChild) {
				domNode.firstChild.nodeValue = val;
			} else {
				domNode = document.createElementNS(this.gd, keys[key]);
				var textNode = document.createTextNode(val);
				domNode.appendChild(textNode);
				sAddrNode.appendChild(domNode);
			}
			empty = false;
		} else if (domNode) {
			sAddrNode.removeChild(domNode);
		}
	}

	if (empty && sAddrNode && !newNode) {
		this.entry.removeChild(sAddrNode);
	} else if (!empty && newNode) {
		this.entry.appendChild(sAddrNode);
	}
}

GCContactEntry.prototype.parseTBAddr = function(abCard, type) 
{
	var fields = new Array();
	var format = nsPreferences.copyUnicharPref("gmcont.addr_conv_format", "");
	var delimiter = nsPreferences.copyUnicharPref("gmcont.addr_conv_delimiter", ", ");
	if (delimiter == '<BR>') delimiter = "\n";
	var isEmpty = true;
	if (format) {
		var items = format.split(",");
		for (var i=0; i<items.length; i++) {
			var value = abCard.getProperty(type+items[i], null);
			if (!value) value = "";
			else isEmpty = false;
			//if (value) fields.push(value);
			fields.push(value);
		}
	}
	return isEmpty ? null : fields.join(delimiter);
}

GCContactEntry.prototype.convertTB2PropertyToTB3 = function(propName)
{
	var cardMaps = [];
	// Contact > Name
  cardMaps["firstName"] = "FirstName";
  cardMaps["phoneticFirstName"] = "PhoneticFirstName";
  cardMaps["lastName"] = "LastName";
  cardMaps["phoneticLastName"] = "PhoneticLastName";
  cardMaps["displayName"] = "DisplayName";
  cardMaps["nickName"] = "NickName";
  // Contact > Internet
  cardMaps["primaryEmail"] = "PrimaryEmail";
  cardMaps["secondEmail"] = "SecondEmail";
  cardMaps["aimScreenName"] = "_AimScreenName"; // NB: AIM.
	cardMaps["preferMailFormat"] = "PreferMailFormat";
	//cardMaps["allowRemoteContent"] = "AllowRemoteContent";
  // Contact > Phones
  cardMaps["workPhone"] = "WorkPhone";
  cardMaps["homePhone"] = "HomePhone";
  cardMaps["faxNumber"] = "FaxNumber";
  cardMaps["pagerNumber"] = "PagerNumber";
  cardMaps["cellularNumber"] = "CellularNumber";
  // Address > Home
  cardMaps["homeAddress"] = "HomeAddress";
  cardMaps["homeAddress2"] = "HomeAddress2";
  cardMaps["homeCity"] = "HomeCity";
  cardMaps["homeState"] = "HomeState";
  cardMaps["homeZipCode"] = "HomeZipCode";
  cardMaps["homeCountry"] = "HomeCountry";
  cardMaps["webPage2"] = "WebPage2";
  // Address > Work
  cardMaps["jobTitle"] = "JobTitle";
  cardMaps["department"] = "Department";
  cardMaps["company"] = "Company";
  cardMaps["workAddress"] = "WorkAddress";
  cardMaps["workAddress2"] = "WorkAddress2";
  cardMaps["workCity"] = "WorkCity";
  cardMaps["workState"] = "WorkState";
  cardMaps["workZipCode"] = "WorkZipCode";
  cardMaps["workCountry"] = "WorkCountry";
  cardMaps["webPage1"] = "WebPage1";
  // Other > (custom)
  cardMaps["custom1"] = "Custom1";
  cardMaps["custom2"] = "Custom2";
  cardMaps["custom3"] = "Custom3";
  cardMaps["custom4"] = "Custom4";
  // Other > Notes
  cardMaps["notes"] = "Notes";

	return cardMaps[propName] ? cardMaps[propName] : propName;
}

GCContactEntry.prototype.dump = function(card)
{
	var cardFields = [
				 // Contact > Name
         "FirstName",
         "PhoneticFirstName",
         "LastName",
         "PhoneticLastName",
         "DisplayName",
         "NickName",
          // Contact > Internet
         "PrimaryEmail",
         "SecondEmail",
         "_GoogleTalk",
         "_AimScreenName",
         "_Yahoo",
         "_Skype",
         "_QQ",
         "_MSN",
         "_ICQ",
         "_JabberId",
    		 "PreferMailFormat",
				 //"AllowRemoteContent",
    		 "AllowRemoteContent",
    		 "PreferDisplayName",
          // Contact > Phones
         "WorkPhone",
         "HomePhone",
         "FaxNumber",
         "PagerNumber",
         "CellularNumber",
          // Address > Home
         "HomeAddress",
         "HomeAddress2",
         "HomeCity",
         "HomeState",
         "HomeZipCode",
         "HomeCountry",
         "WebPage2",
          // Address > Work
         "JobTitle",
         "Department",
         "Company",
         "WorkAddress",
         "WorkAddress2",
         "WorkCity",
         "WorkState",
         "WorkZipCode",
         "WorkCountry",
         "WebPage1",
         //Birthday
         "BirthYear",
         "BirthMonth",
         "BirthDay",
          // Other > (custom)
         "Custom1",
         "Custom2",
         "Custom3",
         "Custom4",
          // Other > Notes
         "Notes"
         ];
  var str = "";
  for (var i=0; i<cardFields.length; i++) {
		if (card.getProperty(cardFields[i], null)) str = str + card.getProperty(cardFields[i], null);
	}
	return str;
	//return card.convertToEscapedVCard()+card.phoneticLastName+card.phoneticFirstName+card.aimScreenName;
}

GCContactEntry.prototype.isChanged = function(card)
{
	var ret = this.dumpstr != this.dump(card);
	return ret;
}

GCContactEntry.prototype.getTitle = function()
{
	if (!this.entry && !this.loadCache()) return "";
	var nameNode = this.entry.getElementsByTagNameNS(this.gd, "fullName")[0];
	return nameNode && nameNode.firstChild ? nameNode.firstChild.nodeValue : "";
}
	
GCContactEntry.prototype.getPhotoLink = function(forAdd)
{

	if (!this.entry && !this.loadCache()) return null;
	var linkNodes = this.entry.getElementsByTagNameNS(this.atom, "link");
	var linkNode = null;
	for (var i=0; i<linkNodes.length; i++) {
		if (linkNodes[i].getAttribute("rel") == "http://schemas.google.com/contacts/2008/rel#photo") {
			linkNode = linkNodes[i];
			break;
		}
	}
	
	var link = null;
	if (linkNode && (forAdd || linkNode.getAttributeNS(this.gd, "etag"))) {
			link = linkNode.getAttribute("href");
	}

	return link;
}

GCContactEntry.prototype.getPhotoEtag = function()
{
	if (!this.entry && !this.loadCache()) return null;
	var linkNodes = this.entry.getElementsByTagNameNS(this.atom, "link");
	var linkNode = null;
	for (var i=0; i<linkNodes.length; i++) {
		if (linkNodes[i].getAttribute("rel") == "http://schemas.google.com/contacts/2008/rel#photo") {
			linkNode = linkNodes[i];
			break;
		}
	}
	
	return linkNode ? linkNode.getAttributeNS(this.gd, "etag") : null;
}

GCContactEntry.prototype.getDispName = function()
{
	return this.getTitle();
}

GCContactEntry.prototype.getPrimaryAddress = function()
{
	if (!this.entry && !this.loadCache()) return null;
	var addrNodes = this.entry.getElementsByTagNameNS(this.gd, "email");
	var addr = "";
	for (var i=0; i<addrNodes.length; i++) {
		if (addrNodes[i].getAttribute("primary") == "true") {
			addr = addrNodes[i].getAttribute("address");
			break;
		}
	}
	
	if (!addr && addrNodes.length > 0) addr = addrNodes[0].getAttribute("address");
	
	if (!addr && this.card) { //pseudo email
		addr = this.card.getProperty("PrimaryEmail", "");
	}
	
	return addr;
}

GCContactEntry.prototype.getAddresses = function()
{
	var ret = new Array();
	if (!this.entry && !this.loadCache()) return ret;
	var addrNodes = this.entry.getElementsByTagNameNS(this.gd, "email");
	for (var i=0; i< addrNodes.length; i++) {
		var addrNode = addrNodes[i];
		var rel = addrNode.getAttribute("rel");
		var relParts = rel ? rel.split("#") : null;
		var addrObj = {
			address: addrNode.getAttribute("address"),
			type: relParts[1],
			primary: addrNode.getAttribute("primary")
		};
		ret.push(addrObj);
	}
	
	return ret;
}

GCContactEntry.prototype.print = function()
{
	gGCLogger.dump("#####\n");
	gGCLogger.dump("Id: " + this.getId()+"\n");
	var addrs = this.getAddresses();
	for (var i=0; i<addrs.length; i++) {
		gGCLogger.dump("Address: " + addrs[i].address+"\n");
		gGCLogger.dump("\tType: " + addrs[i].type+"\n");
		gGCLogger.dump("\tPrimary: " + addrs[i].primary+"\n");
	}
}
